# Author: Carsten Sachse 
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details
"""
Program to assign a series of 3D models to class averages
"""
from EMAN2 import EMData, EMUtil
from collections import OrderedDict, namedtuple
from spring.csinfrastr.csdatabase import SpringDataBase, SegmentTable, base
from spring.csinfrastr.csfeatures import Features
from spring.csinfrastr.cslogger import Logger
from spring.csinfrastr.csproductivity import OpenMpi, Temporary
from spring.csinfrastr.csreadinput import OptHandler
from spring.segment2d.segmentclass import SegmentClass
from spring.segment2d.segmentselect import SegmentSelect
from spring.segment3d.refine.sr3d_main import SegmentRefine3d
from spring.segment3d.refine.sr3d_mpi import SegmentRefine3dMpi
from tabulate import tabulate
import json
import os
import shutil


class SegClassModelPar(object):
    def __init__(self):
        # package/program identity
        self.package = 'emspring'
        self.progname = 'segclassmodel'
        self.proginfo = __doc__
        self.code_files = [self.progname, self.progname + '_mpi']

        self.segclassmodel_features = Features()
        self.feature_set = self.segclassmodel_features.setup(self)
        
        self.define_parameters_and_their_properties()
        self.define_program_states()


    def define_parameters_and_their_properties(self):
        self.feature_set = self.set_inp_multiple_models(self.feature_set)
        self.feature_set = self.segclassmodel_features.set_class_avg_stack(self.feature_set)
        self.feature_set = self.segclassmodel_features.set_spring_path_segments(self.feature_set)
        
        self.feature_set = self.segclassmodel_features.set_in_or_exclude_classes_option(self.feature_set)
        self.feature_set = self.segclassmodel_features.set_in_or_exclude_classes(self.feature_set)
        self.feature_set = self.segclassmodel_features.set_classes_to_be_in_or_excluded(self.feature_set)
        
        self.feature_set = self.segclassmodel_features.set_angular_projection_count(self.feature_set)
        self.feature_set = self.segclassmodel_features.set_out_of_plane_tilt_angle_range(self.feature_set)
        
        self.feature_set = self.segclassmodel_features.set_keep_intermediate_files_option(self.feature_set, 
        hinttxt='Keep reprojection stack of models.')

        self.feature_set = self.segclassmodel_features.set_pixelsize(self.feature_set)
        self.feature_set = self.segclassmodel_features.set_helix_width_and_height(self.feature_set)
        self.feature_set = self.segclassmodel_features.set_alignment_rotation_and_translation(self.feature_set)
        self.feature_set = self.segclassmodel_features.set_internal_binning(self.feature_set)
        
        self.feature_set = self.segclassmodel_features.set_filter_options(self.feature_set)

        self.feature_set = self.segclassmodel_features.set_mpi(self.feature_set)
        self.feature_set = self.segclassmodel_features.set_ncpus(self.feature_set)
        self.feature_set = self.segclassmodel_features.set_temppath(self.feature_set)

        
    def set_inp_multiple_models(self, feature_set):
        inp1 = 'Reference volumes'
        feature_set.parameters[inp1]='recvol???.hdf'
        feature_set.properties[inp1]=feature_set.file_properties(10, ['hdf'], 'getFiles')

        feature_set.hints[inp1]='Input 3D reconstructions: to be reprojected and used for projection matching with ' + \
        'provided classes. They should be generated using segmentrefine3d or segclassreconstruct because they ' + \
        'require \'rise/rotation\' and \'point_sym\' entries in their header.' 
        
        feature_set.level[inp1]='beginner'

        return feature_set
    

    def define_program_states(self):
        self.feature_set.program_states['prepare_models']='Prepare provided models for 3D reprojection'
        self.feature_set.program_states['launch_segmentalign2d_to_match_classes_against_projections']= \
        'Aligns class averages to model reprojections'

        
class SegClassModelPrepare(object):
    def __init__(self, parset = None):
        self.log = Logger()
        if parset is not None:
            self.feature_set = parset
            p = self.feature_set.parameters

            self.infile = p['Class average stack']
            self.models_input = p['Reference volumes']
            self.references = Features().convert_list_of_files_from_entry_string(self.models_input)
            self.spring_path = p['spring.db file']

            self.classes_selection = p['Classes select option']
            self.classes_in_or_exclude = p['Include or exclude classes']
            self.classes_entries = p['Classes list']
        
            self.keep_intermediate_files = p['Keep intermediate files']
            self.pixelsize = p['Pixel size in Angstrom']
            self.helixwidth, self.helixheight = p['Estimated helix width and height in Angstrom']
            self.helixwidthpix = int(round(self.helixwidth / self.pixelsize))
            self.helix_heightpix = int(round(self.helixheight / self.pixelsize))
        
            self.binfactor = p['Internal binning factor']
            self.restrain_inplane_rotation = p['Limit in-plane rotation']
            self.delta_psi = p['Delta in-plane rotation angle']
            if not self.restrain_inplane_rotation:
                self.delta_psi = 180
            self.x_range_A, self.y_range_A = p['X and Y translation range in Angstrom']

            self.out_of_plane_tilt_angle_range = p['Out-of-plane tilt angle range']
        
            self.azimuthal_angle_count, self.out_of_plane_tilt_angle_count = p['Number of projections azimuthal/' + \
            'out-of-plane angle']

            self.high_pass_filter_option = p['High-pass filter option']
            self.low_pass_filter_option = p['Low-pass filter option']
            self.high_pass_filter_cutoff = p['High and low-pass filter cutoffs in 1/Angstrom'][0]
            self.low_pass_filter_cutoff = p['High and low-pass filter cutoffs in 1/Angstrom'][1]
            self.custom_filter_option = p['Custom filter option']
            self.custom_filter_file = p['Custom-built filter file']
            self.bfactor = p['B-Factor']

            self.mpi_option = p['MPI option']
            self.cpu_count = p['Number of CPUs']
            self.temppath=p['Temporary directory']
            

    def prepare_sr3d_object_with_filter_settings(self):
        if hasattr(self, 'comm'):
            sr3d = SegmentRefine3dMpi()
            [setattr(sr3d, each_param, getattr(self, each_param)) for each_param in ['comm', 'rank', 'size']]
        else:
            sr3d = SegmentRefine3d()

        params = ['low_pass_filter_option', 'low_pass_filter_cutoff', 'high_pass_filter_option',
        'high_pass_filter_cutoff', 'custom_filter_option', 'custom_filter_file', 'bfactor', 'pixelsize', 'helixwidth', 
        'helixwidthpix', 'out_of_plane_tilt_angle_range', 'azimuthal_angle_count', 'out_of_plane_tilt_angle_count', 
        'tempdir', 'keep_intermediate_files']

        [setattr(sr3d, each_param, getattr(self, each_param)) for each_param in params]
        
        sr3d.helix_inner_widthpix = 0
        sr3d.layer_line_filter = False
        sr3d.fsc_filter = None

        return sr3d


class SegClassModelProject(SegClassModelPrepare):
    def get_helical_and_point_group_symmetry_from_volume(self, ref_file):
        ref_vol = EMData()
        ref_vol.read_image(ref_file)

        msg = ''
        if not ref_vol.has_attr('rise/rotation') or not ref_vol.has_attr('point_sym'):
            msg += 'Provided model file {0} does not have the \'rise/rotation\' attribute in the header. '
        if not ref_vol.has_attr('point_sym'):
            msg += 'Provided model file {0} does not have the \'point_sym\' attribute in the header. '
        if msg != '':
            msg +='Double-check that it was generated by segmentrefine3d or segclassreconstruct.'
            raise ValueError(msg)

        hel_sym_entry = ref_vol.get_attr('rise/rotation')
        hel_sym = hel_sym_entry.strip('(').strip(')').split(',')
        rise = float(hel_sym[0])
        rotation = float(hel_sym[1])
        helical_symmetry = (rise, rotation)

        point_symmetry = ref_vol.get_attr('point_sym')
        rotational_symmetry = int(point_symmetry[-1])
        if ref_vol.has_attr('apix_z'):
            if ref_vol.get_attr('apix_z') != 1.0:
                pixelsize = ref_vol.get_attr('apix_z')
            else:
                pixelsize = self.pixelsize

        rec_size = ref_vol.get_xsize()

        ref_vol = \
        SegmentRefine3d().rescale_reference_volume_in_case_vol_pixelsize_differs_from_current_pixelsize(rec_size,
        ref_vol, self.pixelsize)

        ref_vol.write_image(os.path.basename(ref_file))

        return rec_size, helical_symmetry, rotational_symmetry, point_symmetry, pixelsize


    def prepare_prj_through_series_of_models(self):
        class_avg = EMData()
        class_avg.read_image(self.infile)
        projection_size = class_avg.get_xsize()
        projection_info = []

        return projection_size, projection_info


    def prepare_volume_for_projection(self, projection_size, sr3d, each_model_id, each_reference):
        recsize, helical_symmetry, rotational_symmetry, point_symmetry, pixelsize = \
        self.get_helical_and_point_group_symmetry_from_volume(each_reference)
        
        pixel_info_nt = sr3d.make_pixel_info_named_tuple()
        pixelinfo = pixel_info_nt(pixelsize, projection_size, recsize, self.helixwidthpix, 0, self.helix_heightpix)
        
#         sr3d.pixelsize = pixelsize
#         sr3d.helical_symmetry = helical_symmetry
#         sr3d.rotational_symmetry_start = int(point_symmetry[-1])
        reference_file = os.path.basename(each_reference)

        reference_info_nt = sr3d.make_reference_info_named_tuple()

        reference_info = reference_info_nt(each_model_id, reference_file, None, None, helical_symmetry, point_symmetry,
        rotational_symmetry, None)

        ref_vol = sr3d.filter_and_mask_reference_volume('medium', reference_info, pixelinfo)

        ref_vol.write_image(reference_file)

        return reference_info, pixelinfo


    def summarize_prj_info(self, projection_info, each_model_id, projection_stack, projection_parameters):
        model_ids = len(projection_parameters) * [each_model_id]
        projection_stack = len(projection_parameters) * [projection_stack]
        prj_ids = list(range(len(projection_parameters)))
        projection_info += zip(prj_ids, model_ids, projection_stack)

        return projection_info

    def get_prj_info_nt(self):
        return namedtuple('prj_info', 'prj_id model_id prj_stack')


    def get_proj_refs_file_name(self):
        merged_prj_stack = 'proj_refs.hdf'

        return merged_prj_stack


    def merge_prj_stacks_and_collect_prj_info(self, projection_info):
        prj = EMData()
        merged_prj_stack = self.get_proj_refs_file_name()
        for each_total_id, (each_prj_id, each_model_id, each_projection_stack) in enumerate(projection_info):
            prj.read_image(each_projection_stack, each_prj_id)
            prj.write_image(merged_prj_stack, each_total_id)
        
        single_prj_files = set(zip(*projection_info)[2])
        [os.remove(each_prj_file) for each_prj_file in single_prj_files]
        
        prj_info = self.get_prj_info_nt()

        projection_info = [prj_info._make(each_info) for each_info in projection_info]

        return merged_prj_stack, projection_info


    def prepare_class_avg_stack(self):
        class_count = EMUtil.get_image_count(self.infile)
        classes = set(range(class_count))
        
        if self.classes_selection:
            classes_sel = SegmentSelect().prepare_list_from_comma_separated_input(self.classes_entries, 'class')
            if self.classes_in_or_exclude == 'include':
                classes = classes.intersection(classes_sel)
            if self.classes_in_or_exclude == 'exclude':
                classes = classes.difference(classes_sel)

        img = EMData()
        local_infile = os.path.splitext(os.path.basename(self.infile))[0] + '_tmp' + \
        os.path.splitext(os.path.basename(self.infile))[-1]
        for each_sel_class, each_class in enumerate(classes):
            img.read_image(self.infile, each_class)
            img.write_image(local_infile, each_sel_class)

        return local_infile, classes
    
                
    def prepare_merged_stack_of_projections(self):
        sr3d = self.prepare_sr3d_object_with_filter_settings()

        projection_size, projection_info = self.prepare_prj_through_series_of_models()

        for each_model_id, each_reference in enumerate(self.references):
            reference_info, pixelinfo = self.prepare_volume_for_projection(projection_size, sr3d, each_model_id,
            each_reference)

            projection_stack, projection_parameters, fine_projection_stack, fine_projection_parameters = \
            sr3d.project_through_reference_volume_in_helical_perspectives('medium', reference_info.model_id,
            reference_info.ref_file, pixelinfo, reference_info.helical_symmetry, reference_info.rotational_symmetry)

            projection_info = self.summarize_prj_info(projection_info, each_model_id, projection_stack,
            projection_parameters)
        
        merged_prj_stack, projection_info = self.merge_prj_stacks_and_collect_prj_info(projection_info)

        return  merged_prj_stack, projection_info


class SegClassModel(SegClassModelProject):
    def launch_segmentalign2d_to_match_classes_against_projections(self, infilestack, merged_prj_stack):
        self.log.fcttolog()
        
        self.aligndict = OrderedDict()
        self.aligndict['Image input stack'] = infilestack
        alistack = os.path.splitext(os.path.basename(infilestack))[0] + 'ali.hdf'
        self.aligndict['Image output stack'] = alistack
        self.aligndict['Number of iterations']= 3
        self.aligndict['Internal binning factor']= 1
        self.aligndict['Update references']=False
        self.aligndict['Aligned average stack'] = 'multi_ref_avg.hdf'
        self.aligndict['Image reference stack'] = merged_prj_stack
        self.aligndict['Reference option'] = True
        self.aligndict['Pixel size in Angstrom'] = self.pixelsize
        
        self.aligndict['Estimated helix width and height in Angstrom'] = ((self.helixwidth, self.helixheight))
        self.aligndict['Limit in-plane rotation']=self.restrain_inplane_rotation
        self.aligndict['Delta in-plane rotation angle'] = self.delta_psi
        self.aligndict['X and Y translation range in Angstrom'] = ((self.x_range_A, self.y_range_A))
    
        self.aligndict['Local refinement'] = False
        self.aligndict['Absolute X and Y translation limit in Angstrom']= ((self.x_range_A, self.y_range_A))
        
        self.aligndict['High-pass filter option']=self.high_pass_filter_option
        self.aligndict['Low-pass filter option']=self.low_pass_filter_option
        self.aligndict['High and low-pass filter cutoffs in 1/Angstrom']=((self.high_pass_filter_cutoff, 
                                                                           self.low_pass_filter_cutoff))
        self.aligndict['Automatic filter option']=False
        self.aligndict['Custom filter option']=self.custom_filter_option
        self.aligndict['Custom-built filter file']=self.custom_filter_file
        self.aligndict['B-Factor']=self.bfactor
        
        self.aligndict['MPI option'] = self.mpi_option
        self.aligndict['Number of CPUs']=self.cpu_count
        self.aligndict['Temporary directory'] = self.temppath 

        segmentalign_dir ='segmentalign2d00' 

        alistack = SegmentClass().run_segmentalign2d_in_separate_process(self.aligndict, segmentalign_dir,
        'segmentalign2d')

        if not self.keep_intermediate_files:
            files_in_dir = os.listdir(segmentalign_dir)
            [os.remove(os.path.join(segmentalign_dir, each_file)) for each_file in files_in_dir]
            os.rmdir(segmentalign_dir)
            os.remove(merged_prj_stack)

        return infilestack
        
    
    def get_matched_prj_id_from_alignment_stack_and_assign_model(self, infilestack, prj_info, reduced_classes):
        class_count = EMUtil.get_image_count(infilestack)
        class_img = EMData()

        cls_assigned_prj = []
        for each_class in list(range(class_count)):
            class_img.read_image(infilestack, each_class)
            cls_assigned_prj.append(class_img.get_attr('assign'))
                         
        cls_assigned_models = [prj_info[each_assigned_prj].model_id for each_assigned_prj in cls_assigned_prj]
    
        os.remove(infilestack)
        all_class_count = EMUtil.get_image_count(self.infile)
        if all_class_count > class_count:
            all_cls_assigned_models = []
            for each_class in list(range(all_class_count)):
                if each_class in reduced_classes:
                    all_cls_assigned_models += [cls_assigned_models[each_red_class_id] 
                                                for each_red_class_id, each_red_class in enumerate(reduced_classes) 
                                                if each_class == each_red_class]
                if each_class not in reduced_classes:
                    all_cls_assigned_models += [None]
        else:
            all_cls_assigned_models = cls_assigned_models
            
        return all_cls_assigned_models
    

    def udpate_segments_in_spring_db_with_assigned_model(self, cls_assigned_models):
        shutil.copy(self.spring_path, 'spring.db')
        session = SpringDataBase().setup_sqlite_db(base)
        
        segments = session.query(SegmentTable).order_by(SegmentTable.id).all()
        
        loginfo =[]
        for each_segment in segments:
            if each_segment.class_id is not None:
                assigned_model = cls_assigned_models[each_segment.class_id]
                each_segment.class_model_id = assigned_model
                session.merge(each_segment)
                loginfo.append((each_segment.stack_id, each_segment.class_id, assigned_model))
            
        session.commit()
        msg = tabulate(loginfo, ['stack_id', 'class_id', 'class_model_id'])
        self.log.ilog('The following segments on the stack have been assigned with the respective models:\n{0}'.format(msg))


    def perform_projection_or_pick_up_completed_prj_in_case_of_mpi(self):
        if not self.mpi_option:
            merged_prj_stack, prj_info = self.prepare_merged_stack_of_projections()
        else:
            merged_prj_stack = self.get_proj_refs_file_name()
            if not os.path.isfile(merged_prj_stack):
                msg = '{0} did not generate projections correctly. '.format(self.feature_set.progname.title())
                'Look for previous error messages to find out the reason.'
                raise ValueError(msg)

            prj_info = json.load(open('prj_info.dat'))
            os.remove('prj_info.dat')
            prj_info_nt = self.get_prj_info_nt()
            prj_info = [prj_info_nt._make(each_prj_info) for each_prj_info in prj_info]

        return merged_prj_stack, prj_info


    def match_reprojections_to_classes(self):
        OpenMpi().setup_and_start_mpi_version_if_demanded(self.mpi_option, self.feature_set, self.cpu_count,
        exiting=False)

        self.tempdir = Temporary().mktmpdir(self.temppath)

        local_infile, reduced_classes = self.prepare_class_avg_stack()
        merged_prj_stack, prj_info = self.perform_projection_or_pick_up_completed_prj_in_case_of_mpi()
            
        infilestack = self.launch_segmentalign2d_to_match_classes_against_projections(local_infile, merged_prj_stack)

        cls_assigned_models = self.get_matched_prj_id_from_alignment_stack_and_assign_model(infilestack, prj_info,
        reduced_classes)
        
        self.udpate_segments_in_spring_db_with_assigned_model(cls_assigned_models)

        os.rmdir(self.tempdir)
        self.log.endlog(self.feature_set)


def main():
    # Option handling
    parset = SegClassModelPar()
    mergeparset = OptHandler(parset)

    ######## Program
    classes = SegClassModel(mergeparset)
    classes.match_reprojections_to_classes()


if __name__ == '__main__':
    main()



# Author: Carsten Sachse 08-Jun-2011
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details
"""
Program to optimize segmentrefine3d reconstruction by varying refinement parameters systematically on a grid
"""

from collections import OrderedDict
from spring.csinfrastr.csdatabase import SpringDataBase, refine_base, RefinementCycleTable, grid_base, GridTable, \
    GridRefineTable
from spring.csinfrastr.csfeatures import Features
from spring.csinfrastr.cslogger import Logger
from spring.csinfrastr.csproductivity import OpenMpi
from spring.csinfrastr.csreadinput import OptHandler
from spring.segment3d.refine.sr3d_main import SegmentRefine3d
from spring.segment3d.refine.sr3d_parameters import SegmentRefine3dPar
from spring.segment3d.refine.sr3d_prepare import SegmentRefine3dPreparation
from spring.segment3d.segclassreconstruct import SegClassReconstruct
from sqlalchemy.sql.expression import desc, asc
from tabulate import tabulate
import os
import shutil


class SegRefine3dGridPar(SegmentRefine3dPar):
    """
    Class to initiate default dictionary with input parameters including help and range values and status dictionary
    """
    def __init__(self):
        # package/program identity
        self.package = 'emspring'       
        self.progname = 'segrefine3dgrid'
        self.proginfo = __doc__
        self.code_files = [self.progname]

        self.segmentrefine3d_features = Features()
        self.feature_set = self.segmentrefine3d_features.setup(self)
        
        self.define_parameters_and_their_properties()
        self.define_program_states()


    def define_parameters_and_their_properties(self):
        super(SegRefine3dGridPar, self).define_input_output_segmentrefine3d()
        parameters_to_be_varied = self.get_parameters_to_be_varied()
        self.feature_set = self.segmentrefine3d_features.set_first_parameter_choice(self.feature_set, parameters_to_be_varied)
        self.feature_set = self.segmentrefine3d_features.set_second_parameter_choice(self.feature_set, parameters_to_be_varied)
        
        self.feature_set = self.segmentrefine3d_features.set_lower_and_upper_limit_first_parameter(self.feature_set)
        self.feature_set = self.segmentrefine3d_features.set_lower_and_upper_limit_second_parameter(self.feature_set)
        self.feature_set = self.segmentrefine3d_features.set_first_and_second_parameter_increment(self.feature_set)
        self.feature_set = self.segmentrefine3d_features.set_subgrid_option(self.feature_set)
        self.feature_set = self.segmentrefine3d_features.set_subgrid_details(self.feature_set)
        self.feature_set = self.segmentrefine3d_features.set_continue_grid_option(self.feature_set)
        self.feature_set = self.segmentrefine3d_features.set_grid_database_cont(self.feature_set)
        
        super(SegRefine3dGridPar, self).define_segmentrefine3d_parameters()
        
    def define_program_states(self):
        super(SegRefine3dGridPar, self).define_program_states()

    def get_parameters_to_be_varied(self):
        parameter_to_be_varied = ['number_of_iterations', 'high_pass_filter_cutoff', 'low_pass_filter_cutoff', 
                                  'b-factor', 'include_classes', 'exclude_curvature_above', 
                                  'exclude_layer_line_ccc_below', 'exclude_defocus_below', 'exclude_defocus_above', 
                                  'exclude_astigmatism_above', 'exclude_projection_ccc_below', 
                                  'exclude_out_of_plane_tilt_below', 'exclude_out_of_plane_tilt_above', 
                                  'exclude_shift_normal_to_helix_above', 'helical_rise_or_pitch', 
                                  'helical_rotation_or_number_of_units_per_turn', 'helix_start', 
                                  'delta_in_plane_rotation', 'out_of_plane_tilt_search_range', 
                                  'out_of_plane_count', 'azimuthal_count', 'alignment_image_size', 'segmentation_step',
                                  'pixelsize']
        return parameter_to_be_varied


class SegRefine3dGridSetup(SegmentRefine3dPreparation):
    def __init__(self, parset = None):
        self.log = Logger()
        if parset is not None:
            self.feature_set = parset
            p = self.feature_set.parameters
            self.define_all_segmentrefine3d_parameters(p)
            
            self.primary_variable = p['First parameter']
            self.secondary_variable = p['Second parameter']
            self.primary_range = p['Lower and upper limit first parameter']
            self.primary_inc, self.second_inc = p['First and second parameter increment']
            self.secondary_range = p['Lower and upper limit second parameter']
            self.subgrid_option = p['Subgrid option']
            self.subgrid = p['Part and number of subgrids']
            self.continue_grid = p['Grid continue option']
            self.grid_database = p['Grid database']
            
        
    def generate_segmentrefine3d_parameters_while_removing_grid_parameters(self, parameters):
        
        segmentrefine3d_params = OrderedDict()
        for each_param in parameters:
            if each_param not in ['First parameter', 'Second parameter', 'Lower and upper limit first parameter',
                                  'Lower and upper limit second parameter', 'First and second parameter increment',
                                  'Subgrid option', 'Part and number of subgrids', 'Grid continue option', 
                                  'Grid database']:
                segmentrefine3d_params[each_param] = parameters[each_param]
            
        return segmentrefine3d_params
    

    def setup_non_refinement_selection_criteria(self, each_grid_point, parameters, index, one_of_both_parameters):
        if one_of_both_parameters == 'include_classes':
            parameters['Classes select option'] = True
            parameters['Include or exclude classes'] = 'include'
            parameters['Classes list'] = each_grid_point[index]
        elif one_of_both_parameters == 'exclude_curvature_above':
            parameters['Straightness select option'] = True
            parameters['Include or exclude straight helices'] = 'include'
            parameters['Persistence length range'] = (0, each_grid_point[index])
        elif one_of_both_parameters == 'exclude_layer_line_ccc_below':
            parameters['Layer line correlation select option'] = True
            parameters['Include or exclude segments based on layer-line correlation'] = 'exclude'
            parameters['Correlation layer line range'] = (0, each_grid_point[index])
        elif one_of_both_parameters == 'exclude_defocus_below':
            parameters['Defocus select option'] = True
            parameters['Include or exclude defocus range'] = 'exclude'
            parameters['Defocus range'] = (0, each_grid_point[index])
        elif one_of_both_parameters == 'exclude_defocus_above':
            parameters['Defocus select option'] = True
            parameters['Include or exclude defocus range'] = 'include'
            parameters['Defocus range'] = (0, each_grid_point[index])
        elif one_of_both_parameters == 'exclude_astigmatism_above':
            parameters['Astigmatism select option'] = True
            parameters['Include or exclude astigmatic segments'] = 'include'
            parameters['Astigmatism range'] = (0, each_grid_point[index])
            
        return parameters
    

    def setup_refinement_selection_criteria(self, each_grid_point, parameters, index, one_of_both_parameters):
        if one_of_both_parameters == 'exclude_projection_ccc_below':
            parameters['Projection correlation select option'] = True
            parameters['Include or exclude segments based on projection correlation'] = 'exclude'
            parameters['Correlation projection range'] = (0, each_grid_point[index])
        elif one_of_both_parameters == 'exclude_out_of_plane_tilt_below':
            parameters['Out-of-plane tilt select option'] = True
            parameters['Include or exclude out-of-plane tilted segments'] = 'exclude'
            parameters['Out-of-plane tilt range'] = (-each_grid_point[index], each_grid_point[index])
        elif one_of_both_parameters == 'exclude_out_of_plane_tilt_above':
            parameters['Out-of-plane tilt select option'] = True
            parameters['Include or exclude out-of-plane tilted segments'] = 'include'
            parameters['Out-of-plane tilt range'] = (-each_grid_point[index], each_grid_point[index])
        elif one_of_both_parameters == 'exclude_shift_normal_to_helix_above':
            parameters['Shift normal to helix select option'] = True
            parameters['Include or exclude segments with shift normal to helix'] = 'include'
            parameters['Shift normal to helix in Angstrom'] = each_grid_point[index]
        
        return parameters
    

    def setup_symmetry_related_parameters(self, each_grid_point, parameters, index, one_of_both_parameters):
        if one_of_both_parameters == 'helical_rise_or_pitch':
            parameters['Helical symmetry in Angstrom or degrees'] = (each_grid_point[index], 
            parameters['Helical symmetry in Angstrom or degrees'][1])
        elif one_of_both_parameters == 'helical_rotation_or_number_of_units_per_turn':
            pitch =  parameters['Helical symmetry in Angstrom or degrees'][0]
            parameters['Helical symmetry in Angstrom or degrees'] = (pitch, each_grid_point[index])
        elif one_of_both_parameters == 'segmentation_step':
            parameters['Step size of segmentation in Angstrom'] = each_grid_point[index]
        elif one_of_both_parameters == 'alignment_image_size':
            parameters['Image alignment size in Angstrom'] = each_grid_point[index]
        elif one_of_both_parameters == 'pixelsize':
            parameters['Pixel size in Angstrom'] = each_grid_point[index]
        
        return parameters
    

    def setup_alignment_related_parameters(self, each_grid_point, parameters, index, one_of_both_parameters):
        if one_of_both_parameters == 'delta_in_plane_rotation':
            parameters['Delta in-plane rotation angle'] = each_grid_point[index]
        elif one_of_both_parameters == 'out_of_plane_tilt_search_range':
            parameters['Out-of-plane tilt angle range'] = (-each_grid_point[index], each_grid_point[index])
        elif one_of_both_parameters == 'out_of_plane_count':
            parameters['Number of projections azimuthal/out-of-plane angle'] = \
            (parameters['Number of projections azimuthal/out-of-plane angle'][0], each_grid_point[index])
        elif one_of_both_parameters == 'azimuthal_count':
            parameters['Number of projections azimuthal/out-of-plane angle'] = \
            (each_grid_point[index], parameters['Number of projections azimuthal/out-of-plane angle'][1])
            
        return parameters
    

    def setup_filter_related_parameters(self, each_grid_point, parameters, index, one_of_both_parameters):
        if one_of_both_parameters == 'number_of_iterations':
            parameters['Number of iteration'] = each_grid_point[index]
        elif one_of_both_parameters == 'high_pass_filter_cutoff':
            parameters['High and low-pass filter cutoffs in 1/Angstrom'] = \
            (each_grid_point[index], parameters['High and low-pass filter cutoffs in 1/Angstrom'][1])
        elif one_of_both_parameters == 'low_pass_filter_cutoff':
            parameters['High and low-pass filter cutoffs in 1/Angstrom'] = \
            (parameters['High and low-pass filter cutoffs in 1/Angstrom'][0], each_grid_point[index])
        elif one_of_both_parameters == 'b-factor':
            parameters['B-Factor'] = each_grid_point[index]
            
        return parameters
    

    def add_grid_point_to_parameter_dictionary(self, first_param, second_param, each_grid_point, parameters):
        """
        >>> from spring.segment3d.segrefine3dgrid import SegRefine3dGrid
        >>> s = SegRefine3dGrid()
        >>> r_p = 'helical_rise_or_pitch'
        >>> params = SegRefine3dGridPar().parameters
        >>> s.add_grid_point_to_parameter_dictionary('b-factor', r_p, [-111, 1.111], params) #doctest: +NORMALIZE_WHITESPACE
        OrderedDict([('Image input stack refinement', 'protein_stack.hdf'), 
        ('Output volume name', 'recvol.hdf'), 
        ('First parameter', 'helical_rise_or_pitch'), ('Second parameter', 
        'helical_rotation_or_number_of_units_per_turn'), 
        ('Lower and upper limit first parameter', (1.4, 1.9)), 
        ('Lower and upper limit second parameter', (22.0, 24.0)), 
        ('First and second parameter increment', (0.1, 0.3)), 
        ('Subgrid option', False), ('Part and number of subgrids', (1, 3)), 
        ('Grid continue option', False), ('Grid database', 'grid.db'), 
        ('Diagnostic plot prefix', 'diagnostic_plot.pdf'), 
        ('Number of iterations', 20), ('Reference structure option', False), 
        ('Reference volume', 'reference_vol.hdf'), ('spring.db file', 
        'spring.db'), ('Continue refinement option', False), 
        ('refinement.db file', 'refinement.db'), ('Assemble refinement strategy', 
        True), ('LR - Low resolution aim', True), 
        ('LR - azimuthal and out-of-plane search restraint in degrees', 
        (180.0, 180.0)), ('LR - X and Y translation range in Angstrom', 
        (50, 23)), ('MR - Medium resolution aim', True), 
        ('MR - azimuthal and out-of-plane search restraint in degrees', 
        (180.0, 180.0)), ('MR - X and Y translation range in Angstrom', 
        (21, 10)), ('HR - High resolution aim', True), 
        ('HR - azimuthal and out-of-plane search restraint in degrees', 
        (20.0, 20.0)), ('HR - X and Y translation range in Angstrom', 
        (14, 7)), ('MaxR - Maximum resolution aim', True), 
        ('MaxR - azimuthal and out-of-plane search restraint in degrees', 
        (2.0, 2.0)), ('MaxR - X and Y translation range in Angstrom', 
        (7, 3.5)), ('Absolute X and Y translation limit in Angstrom', 
        (100, 100)), ('Frame motion correction', False), 
        ('Frame average window size', 3),  ('Frame local averaging distance', 700), 
        ('Independent half-set refinement', False), 
        ('Half-set refinement start', 'medium'), ('High-pass filter option', False), 
        ('Low-pass filter option', True), 
        ('High and low-pass filter cutoffs in 1/Angstrom', 
        (0.001, 0.09)), ('B-Factor', -111), ('Custom filter option', 
        False), ('Custom-built filter file', 'filter_function.dat'), 
        ('Automatic FSC filter', True), ('Filter layer-lines option', False), 
        ('Micrographs select option', False), ('Include or exclude micrographs', 
        'include'), ('Micrographs list', '1-9, 11, 13'), ('Helices select option', 
        False), ('Include or exclude helices', 'include'), ('Helices list', 
        '1-9, 11, 13'), ('Segments select option', False), 
        ('Include or exclude segments', 'include'), ('Segment file', 
        'stackid_file.dat'), ('Classes select option', False), 
        ('Include or exclude classes', 'include'), ('Class type', 'class_id'), 
        ('Classes list', '1-9, 11, 13'),  ('Persistence class option', False), 
        ('Persistence class length in Angstrom', 700), 
        ('Class occupancy threshold', 0.5), ('Straightness select option', False), 
        ('Include or exclude straight helices', 'include'), ('Persistence length range', 
        (80, 100)), ('Layer line correlation select option', False), 
        ('Include or exclude segments based on layer-line correlation', 
        'include'), ('Correlation layer line range', (60, 100)), 
        ('Defocus select option', False), ('Include or exclude defocus range', 
        'include'), ('Defocus range', (10000, 40000)), 
        ('Astigmatism select option', False), 
        ('Include or exclude astigmatic segments', 'include'), 
        ('Astigmatism range', (0, 4000)), 
        ('Projection correlation select option', False), 
        ('Include or exclude segments based on projection correlation', 
        'include'), ('Correlation projection range', (60, 100)), 
        ('Out-of-plane tilt select option', False), 
        ('Include or exclude out-of-plane tilted segments', 'include'), 
        ('Out-of-plane tilt range', (-5, 5)), 
        ('Shift normal to helix select option', False), 
        ('Include or exclude segments with shift normal to helix', 'include'), 
        ('Shift normal to helix in Angstrom', 5.0), ('Keep intermediate files', 
        False), ('Estimated helix inner and outer diameter in Angstrom', (0, 190)), 
        ('Pixel size in Angstrom', 1.163), ('Symmetrize helix', True), 
        ('Helical rise/rotation or pitch/number of units per turn choice', 
        'rise/rotation'), ('Enforce even phi option', False), 
        ('Release cycle even phi', 8), ('Pitch enforce even phi', 8.0), 
        ('Bin cutoff of phi angles', 100), 
        ('Helical symmetry in Angstrom or degrees', (1.111, 22.03)), 
        ('Rotational symmetry', 1), ('Helix polarity', 'polar'), 
        ('Unbending option', False), ('Force helical continuity', True), 
        ('Limit in-plane rotation', True), ('Delta in-plane rotation angle', 
        10.0), ('Out-of-plane tilt angle range', (-12, 12)), 
        ('Number of projections azimuthal/out-of-plane angle', 
        (90, 7)), ('Image alignment size in Angstrom', 700), 
        ('Step size of segmentation in Angstrom', 70), 
        ('Choose out-of-plane tilt amplitude correlation', False), 
        ('Amplitude correlation out-of-plane tilt range', (-12, 12)), 
        ('3D CTF correction', True), ('3D CTF correction intensity', 'low'), 
        ('MPI option', True), ('Number of CPUs', 8), ('Temporary directory', '/tmp')])
        """
        for index, one_of_both_parameters in enumerate([first_param, second_param]):
            if one_of_both_parameters is not None:
                parameters = self.setup_filter_related_parameters(each_grid_point, parameters, index,
                one_of_both_parameters)
                
                parameters = self.setup_non_refinement_selection_criteria(each_grid_point, parameters, index,
                one_of_both_parameters)
                
                parameters = self.setup_refinement_selection_criteria(each_grid_point, parameters, index,
                one_of_both_parameters)
                
                parameters = self.setup_symmetry_related_parameters(each_grid_point, parameters, index,
                one_of_both_parameters)
                
                parameters = self.setup_alignment_related_parameters(each_grid_point, parameters, index,
                one_of_both_parameters)
        
        return parameters


    def generate_file_name_for_grid_search(self, first_param, second_param, each_grid_point):
        """
        >>> from spring.segment3d.segrefine3dgrid import SegRefine3dGrid
        >>> s = SegRefine3dGrid()
        >>> s.generate_file_name_for_grid_search('b-factor', None, [-200, 3])
        'b-factor_-200'
        >>> sec_par = 'helical_rise_or_pitch'
        >>> s.generate_file_name_for_grid_search('b-factor', sec_par, [-200, 3.344])
        'b-factor_-200_helical_rise_or_pitch_3344'
        """
        if each_grid_point[0] == int(each_grid_point[0]):
            first_grid_point_string = str(int(each_grid_point[0]))
        else:
            first_grid_point_string = ''.join(str(each_grid_point[0]).split('.'))
            
        if each_grid_point[1] == int(each_grid_point[1]):
            second_grid_point_string = str(int(each_grid_point[1]))
        else:
            second_grid_point_string = ''.join(str(each_grid_point[1]).split('.'))
            
        if first_param == second_param or second_param is None:
            run_name = '{0}_{1}'.format(first_param, first_grid_point_string)
        else:
            run_name = '{0}_{1}_{2}_{3}'.format(first_param, first_grid_point_string, second_param,
            second_grid_point_string)
            
        return run_name
    

    def setup_parameters_for_each_grid_point(self, first_param, second_param, each_grid_point, parameters):
        parameters = self.add_grid_point_to_parameter_dictionary(first_param, second_param, each_grid_point, parameters)
        run_name = self.generate_file_name_for_grid_search(first_param, second_param, each_grid_point)
                
        parameter_file = Features().write_parfile(parameters, run_name, tag=False)
        
        return parameter_file, run_name
    
    
class SegRefine3dGrid(SegRefine3dGridSetup):
    
    def split_grid_if_demanded(self, subgrid_option, subgrid, grid_sequence):
        """
        >>> from spring.segment3d.segrefine3dgrid import SegRefine3dGrid
        >>> from spring.segment3d.segclassreconstruct import SegClassReconstruct
        >>> s = SegClassReconstruct()
        >>> grid = s.generate_rise_rotation_or_pitch_unitnumber_pairs_for_symmetry_grid((275.0, 305.0), 5.0, (3.8, 4.2), 0.1)
        >>> SegRefine3dGrid().split_grid_if_demanded(True, (1, 6), grid.ravel())
        array([(275.0, 3.8), (275.0, 3.9), (275.0, 4.0), (275.0, 4.1),
               (275.0, 4.2), (280.0, 3.8)], dtype=object)
        >>> SegRefine3dGrid().split_grid_if_demanded(True, (7, 6), grid.ravel())
        Traceback (most recent call last):
          File '/System/Library/Frameworks/Python.framework/Versions/2.7/lib/python2.7/doctest.py', line 1254, in __run
            compileflags, 1) in test.globs
          File '<doctest segrefine3dgrid.SegRefine3dGrid.split_grid_if_demanded[5]>', line 1, in <module>
            SegRefine3dGrid().split_grid_if_demanded(True, (7, 6), grid.ravel())
          File 'spring/segment3d/segrefine3dgrid.py', line 430, in split_grid_if_demanded
            raise ValueError(msg)
        ValueError: Part of subgrids must not be larger than entire number of subgrids. Enter part of subgrids < number of subgrids.
        """
        if subgrid_option:
            subgrid_id, subgrid_count = subgrid
            if subgrid_id > subgrid_count:
                msg = 'Part of subgrids must not be larger than entire number of subgrids. Enter part of subgrids < ' +\
                'number of subgrids.'
                raise ValueError(msg)
            grid_sequence = OpenMpi().split_sequence_evenly(grid_sequence, subgrid_count)
            grid_sequence = grid_sequence[subgrid_id - 1]
            
        return grid_sequence
    

    def get_grid_parameters_and_final_grid_entry(self, grid_database, grid_name):
        shutil.copy(grid_database, grid_name)
        
        ref_session = SpringDataBase().setup_sqlite_db(grid_base, grid_name)
        ref_grid = ref_session.query(GridTable).first()
        
        last_grid_cycle = ref_session.query(GridRefineTable).order_by(asc(GridRefineTable.id)).all()
        if last_grid_cycle == []:
            last_grid_cycle_id = 0
        else:
            last_grid_cycle_id = last_grid_cycle[-1].id
            columns = SpringDataBase().get_columns_from_table(GridRefineTable)
            log_str = 'The following grid data was read from the previous grid:'
            for each_cycle in last_grid_cycle:
                col_data = [str(getattr(each_cycle, column)) for column in columns]
                msg = tabulate(zip(columns, col_data))
                log_str += '\ngrid_cycle{0}:\n{1}'.format(each_cycle.id, msg)
            
            self.log.ilog(log_str)
        
        return last_grid_cycle_id, ref_grid
    
    
    def get_grid_database_name(self, subgrid_option):
        if subgrid_option:
            subgrid_id, subgrid_count = self.subgrid
            grid_name = 'grid_{0:02}_{1:02}.db'.format(subgrid_id, subgrid_count)
        else:
            grid_name = 'grid.db'

        return grid_name


    def get_variables_ranges_and_increments(self, grid_name):
        if self.continue_grid:
            last_grid_cycle_id, ref_grid = self.get_grid_parameters_and_final_grid_entry(self.grid_database, grid_name)
            if ref_grid.primary_variable in ['Helical rise in Angstrom', 'Helical pitch in Angstrom']:
                self.rise_rot_or_pitch_unit_choice, primary_variable, secondary_variable = \
                SegClassReconstruct().get_correct_primary_and_secondary_variables_from_database(ref_grid.primary_variable, 
                ref_grid.secondary_variable)
            else:
                primary_variable = ref_grid.primary_variable
                secondary_variable = ref_grid.secondary_variable
            primary_range = ref_grid.primary_min, ref_grid.primary_max
            primary_inc = ref_grid.primary_inc
            secondary_range = ref_grid.second_min, ref_grid.second_max
            second_inc = ref_grid.second_inc
        else:
            last_grid_cycle_id = 0
            if self.secondary_variable == 'helical_rise_or_pitch':
                primary_variable = self.secondary_variable
                primary_range = self.secondary_range
                primary_inc = self.second_inc
                secondary_variable = self.primary_variable
                secondary_range = self.primary_range
                second_inc = self.primary_inc
            else:
                primary_variable = self.primary_variable
                primary_range = self.primary_range
                primary_inc = self.primary_inc
                secondary_variable = self.secondary_variable
                secondary_range = self.secondary_range
                second_inc = self.second_inc

            if primary_variable == secondary_variable or secondary_variable == 'none':
                secondary_variable = None
                secondary_range = (None, None)
                second_inc = None
                secondary_range = self.iteration_count, self.iteration_count
                second_inc = 0

        return primary_range, primary_inc, secondary_range, second_inc, last_grid_cycle_id, primary_variable, \
        secondary_variable


    def setup_grid_sequence(self):
        grid_name = self.get_grid_database_name(self.subgrid_option)
            
        primary_range, primary_inc, secondary_range, second_inc, last_grid_cycle_id, primary_variable, \
        secondary_variable = self.get_variables_ranges_and_increments(grid_name)
            
        grid_sequence = \
        SegClassReconstruct().generate_rise_rotation_or_pitch_unitnumber_pairs_for_symmetry_grid(primary_range,
        primary_inc, secondary_range, second_inc)
            
        grid_sequence = grid_sequence.ravel()
        grid_sequence = self.split_grid_if_demanded(self.subgrid_option, self.subgrid, grid_sequence)
        
        grid_str = [(each_id, each_first, each_second) for each_id, (each_first, each_second) in enumerate(grid_sequence)]
        msg = tabulate(grid_str, ['local_grid_id', primary_variable, secondary_variable])
        self.log.ilog('The following grid combinations will be run:\n{0}'.format(msg))
            
        return grid_sequence, last_grid_cycle_id, primary_variable, secondary_variable, grid_name


    def generate_cmdline_for_each_grid_point(self, parameter_file):
        """
        >>> from spring.segment3d.segrefine3dgrid import SegRefine3dGrid
        >>> s = SegRefine3dGrid()
        >>> s.generate_cmdline_for_each_grid_point('rot_23432_rise_4224.par')
        'segmentrefine3d --d rot_23432_rise_4224 --f rot_23432_rise_4224.par'
        """
        cmdline = 'segmentrefine3d --d {0} --f {1}'.format(os.path.splitext(parameter_file)[0],
        parameter_file)
        
        return cmdline
    

    def make_database_entries(self, run_name, each_grid_point, grid_session, last_cycle, stat_file, reproj_file,
    latest_reconstruction):
        grid_cycle = GridRefineTable()
        rundir_name = os.path.abspath(run_name)
        grid_cycle.dirname = rundir_name
        grid_cycle.primary_value = each_grid_point[0]
        grid_cycle.secondary_value = each_grid_point[1]
        grid_cycle.fsc_0143 = last_cycle.fsc_0143
        grid_cycle.fsc_05 = last_cycle.fsc_05
        grid_cycle.helical_ccc_error = last_cycle.helical_ccc_error
        grid_cycle.mean_helical_ccc = last_cycle.mean_helical_ccc
        grid_cycle.out_of_plane_dev = last_cycle.out_of_plane_dev
        grid_cycle.amp_correlation = last_cycle.amp_correlation
        grid_cycle.amp_corr_quarter_nyquist = last_cycle.amp_corr_quarter_nyquist
        grid_cycle.amp_corr_half_nyquist = last_cycle.amp_corr_half_nyquist
        grid_cycle.amp_corr_3quarter_nyquist = last_cycle.amp_corr_3quarter_nyquist
        grid_cycle.variance = last_cycle.variance
        grid_cycle.xshift_error = last_cycle.xshift_error
        grid_cycle.inplane_error = last_cycle.inplane_error
        grid_cycle.outofplane_error = last_cycle.outofplane_error
        grid_cycle.mean_peak = last_cycle.mean_peak
        grid_cycle.excluded_inplane_ratio = last_cycle.excluded_inplane_count / float(last_cycle.segment_count)
        grid_cycle.excluded_out_of_plane_ratio = last_cycle.excluded_out_of_plane_tilt_count / float(last_cycle.segment_count)
        grid_cycle.excluded_total_ratio = last_cycle.total_excluded_count / float(last_cycle.segment_count)
        grid_cycle.asym_unit_count = last_cycle.asym_unit_count
        grid_cycle.avg_azimuth_sampling = last_cycle.avg_azimuth_sampling
        grid_cycle.dev_azimuth_sampling = last_cycle.dev_azimuth_sampling
        grid_cycle.em_files_2d = [(os.path.join(rundir_name, stat_file), 0), (os.path.join(rundir_name, reproj_file), 0)]
        grid_cycle.em_files_3d = [os.path.join(rundir_name, latest_reconstruction)]
        grid_session.add(grid_cycle)
        
        last_grid_entry = grid_session.query(GridTable).order_by(desc(GridTable.id)).first()
        last_grid_entry.completed_grid_id = grid_cycle.id
        
        grid_session.merge(last_grid_entry)
        grid_session.commit()
        
        return grid_cycle
        

    def enter_parameters_in_grid_database(self, primary_variable, secondary_variable, run_name, each_grid_point,
    grid_name, target_cycle_id):
        grid_session = SpringDataBase().setup_sqlite_db(grid_base, grid_name)
                
        ref_db = os.path.join(run_name, 'refinement{0:03}{1}db'.format(target_cycle_id, os.extsep))
        if not os.path.exists(ref_db):
            msg = 'The database file {0} could not be found. This indicates that the refinement '.format(ref_db) + \
            'of the run ({0}) did not complete successfully. Please check the local {1} '.format(run_name, 
            os.path.join(run_name, 'report.log')) + 'file for further error details.'
            raise RuntimeError(msg)
        
        ref_session = SpringDataBase().setup_sqlite_db(refine_base, ref_db)
        
        last_cycle = ref_session.query(RefinementCycleTable).order_by(desc(RefinementCycleTable.id)).first()
        
        stat_file = SegmentRefine3d().generate_diagnostics_statistics_file_name(last_cycle.id, last_cycle.pixelsize,
        self.diagnostic_plot_prefix)
        
        reproj_file = SegmentRefine3d().generate_diagnostics_reprojection_file_name(last_cycle.id, last_cycle.pixelsize,
        self.diagnostic_plot_prefix)
        
        latest_reconstruction = SegmentRefine3d().generate_file_name_with_apix(target_cycle_id, self.outfile_prefix,
        last_cycle.pixelsize)

        this_cycle = self.make_database_entries(run_name, each_grid_point, grid_session, last_cycle, stat_file,
        reproj_file, latest_reconstruction)
        
        columns = SpringDataBase().get_columns_from_table(GridRefineTable)
        log_entries = [str(getattr(this_cycle, each_col)) for each_col in columns]
                                                         
        msg = tabulate(zip(columns, log_entries))

        self.log.ilog('The following parameters have been determined and entered into the database:\n{0}'.format(msg))
        
    
    def enter_variables_into_grid_database(self, primary_variable, secondary_variable, primary_range, secondary_range,
    primary_inc, second_inc, rise_rot_or_pitch_unit_choice):
        primary_db_variable, secondary_db_variable = \
        SegClassReconstruct().get_correct_primary_and_secondary_variables_for_database(rise_rot_or_pitch_unit_choice, 
        secondary_variable)
        
        this_grid_run = SegClassReconstruct().enter_starting_parameters_of_grid_search(primary_db_variable, 
            secondary_db_variable, primary_range, primary_inc, secondary_range, second_inc)
        
        return this_grid_run, primary_db_variable, secondary_db_variable
    

    def prepare_grid_database(self, primary_variable, secondary_variable, primary_range, secondary_range, primary_inc,
    second_inc, grid_name):
        grid_session = SpringDataBase().setup_sqlite_db(grid_base, grid_name)
        this_grid_run, primary_db_variable, secondary_db_variable = self.enter_variables_into_grid_database(primary_variable, 
            secondary_variable, primary_range, secondary_range, primary_inc, second_inc, self.rise_rot_or_pitch_unit_choice)
        
        grid_session.add(this_grid_run)
        grid_session.commit()

        if self.resume_refinement_option:
            cont_ref_session = SpringDataBase().setup_sqlite_db(refine_base, self.refinementdb_path)
            res_cycle = cont_ref_session.query(RefinementCycleTable).order_by(desc(RefinementCycleTable.id)).first()
            if res_cycle is None:
                msg = 'Specified refinement.db does not contain any cycles to be continued. ' + \
                'Please, double-check file {0}'.format(self.refinementdb_path)
                raise ValueError(msg)

            target_cycle_id = res_cycle.id + self.iteration_count
        else:
            target_cycle_id = self.iteration_count

        return primary_db_variable, secondary_db_variable, target_cycle_id


    def launch_segmentrefine3d_jobs(self):
        grid_sequence, last_grid_cycle_id, primary_variable, secondary_variable, grid_name = self.setup_grid_sequence()
        
        segmentrefine3d_parameters = \
        self.generate_segmentrefine3d_parameters_while_removing_grid_parameters(self.feature_set.parameters)
        
        primary_db_variable, secondary_db_variable, target_cycle_id = self.prepare_grid_database(primary_variable,
        secondary_variable, self.primary_range, self.secondary_range, self.primary_inc, self.second_inc, grid_name)
            
        for each_grid_id, each_grid_point in enumerate(grid_sequence):
            if each_grid_id >= last_grid_cycle_id:
                parameter_file, run_name = self.setup_parameters_for_each_grid_point(primary_variable, secondary_variable,
                each_grid_point, segmentrefine3d_parameters)
                
                cmdline = self.generate_cmdline_for_each_grid_point(parameter_file)
                self.log.in_progress_log()
                OpenMpi().launch_command(cmdline)
                
                self.enter_parameters_in_grid_database(primary_db_variable, secondary_db_variable, run_name, 
                each_grid_point, grid_name, target_cycle_id)
            
            self.log.plog(100 * (each_grid_id + 1) / len(grid_sequence))
        
        self.log.endlog(self.feature_set)
            
            
def main():
    # Option handling
    parset = SegRefine3dGridPar()
    mergeparset = OptHandler(parset)

    ######## Program
    stack = SegRefine3dGrid(mergeparset)
    stack.launch_segmentrefine3d_jobs()

if __name__ == '__main__':
    main()


# Author: Carsten Sachse 11-Aug-2011
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details

from spring.csinfrastr.csdatabase import SpringDataBase, base
from spring.csinfrastr.csfeatures import Features
from spring.csinfrastr.csproductivity import OpenMpi
from spring.micprgs.micctfdetermine import MicCtfDetermine, MicCtfDeterminePar
from tabulate import tabulate
import os
import shutil

class ScanMpi(object):
    def startup_scan_mpi_programs(self):
        self.comm, self.rank, self.size, self.log, self.tempdir = OpenMpi().setup_mpi_and_simultaneous_logging(self.log,
        self.feature_set.logfile, self.temppath)
        
        self.outfiles = Features().rename_series_of_output_files(self.micrograph_files, self.outfile)
        
        self.log.fcttolog()
        if self.rank == 0:
            if len(self.micrograph_files) == 0:
                msg = 'No micrographs found in {0}.'.format(self.infile)
                raise ValueError(msg)
            elif len(self.micrograph_files) < self.size:
                msg = 'You requested a larger number of CPUs than microcraphs. To optimally make use of the ' + \
                'resources. Please reduce number of requested CPUs to ' + \
                '{0} for {0} micrographs.'.format(len(self.micrograph_files))
                raise ValueError(msg)
            self.micrograph_files = OpenMpi().split_sequence_evenly(self.micrograph_files, self.size)
            self.outfiles = OpenMpi().split_sequence_evenly(self.outfiles, self.size)
            
            table_data = [ (each_rank_id, ', '.join(each_fileset)) 
                          for each_rank_id, each_fileset in enumerate(self.micrograph_files)]

            msg = tabulate(table_data, ['node_id', 'micrographs'])
            log_str = 'The following nodes will handle the following micrographs:\n{0}'.format(msg)
            self.log.ilog(log_str)
            
        else:
            self.micrograph_files = None
        self.micrograph_files = self.comm.scatter(self.micrograph_files, root=0)
        self.outfiles = self.comm.scatter(self.outfiles, root=0)


    def end_scan_mpi_programs(self):
        self.comm.barrier()
        if self.rank == 0:
            self.log.endlog(self.feature_set)
        os.rmdir(self.tempdir)


class MicCtfDetermineMpi(MicCtfDetermine, ScanMpi):

    def gather_ctf_and_enter_in_database(self, ctffind_params_list, ctftilt_params_list):
        ctffinds = OpenMpi().convert_list_of_namedtuples_to_list_of_lists(ctffind_params_list)
        ctftilts = OpenMpi().convert_list_of_namedtuples_to_list_of_lists(ctftilt_params_list)
        ctffinds = self.comm.gather(ctffinds, root=0)
        ctftilts = self.comm.gather(ctftilts, root=0)
        micrograph_files = self.comm.gather(self.micrograph_files, root=0)
        if self.rank == 0:
            ctffinds = OpenMpi().merge_sequence_of_sequences(ctffinds)
            ctftilts = OpenMpi().merge_sequence_of_sequences(ctftilts)
            micrograph_files = OpenMpi().merge_sequence_of_sequences(micrograph_files)
            ctffind_tuple = self.make_ctffind_parameters_named_tuple()
            
            ctffind_params_combined_list = OpenMpi().convert_list_of_lists_to_list_of_provided_namedtuple(ctffinds,
            ctffind_tuple)
            
            ctftilt_tuple = self.make_ctftilt_parameters_named_tuple()
            
            ctftilt_params_combined_list = OpenMpi().convert_list_of_lists_to_list_of_provided_namedtuple(ctftilts,
            ctftilt_tuple)
            
            self.enter_ctffind_and_ctftilt_values_in_database(micrograph_files, ctffind_params_combined_list,
            ctftilt_params_combined_list)


    def fill_micrographs_list_with_dummy(self, micrograph_files, max_micrograph_count):
        """
        >>> from spring.micprgs.micctfdetermine_mpi import MicCtfDetermineMpi
        >>> MicCtfDetermineMpi().fill_micrographs_list_with_dummy(['dim', 'dum'], 4) 
        ['dim', 'dum', 'place_holder', 'place_holder']
        >>> MicCtfDetermineMpi().fill_micrographs_list_with_dummy(['dim', 'dum'], 2) 
        ['dim', 'dum']
        """
        while len(micrograph_files) < max_micrograph_count:
            micrograph_files += ['place_holder']
        
        return micrograph_files


    def insure_that_every_node_has_the_same_number_of_micrographs(self, micrograph_files):
        micrograph_count = self.comm.gather(len(micrograph_files), root=0)
        if self.rank == 0:
            max_micrograph_count = max(micrograph_count)
        else:
            max_micrograph_count = None
            
        max_micrograph_count = self.comm.bcast(max_micrograph_count, root=0)
        
        micrograph_files = self.fill_micrographs_list_with_dummy(micrograph_files, max_micrograph_count)
        
        return micrograph_files
        
        
    def run_ctffind_and_ctftilt_for_given_micrographs(self, micrograph_files, outfiles):
        if self.rank == 0:
            if self.spring_db_option:
                shutil.copy(self.spring_path, 'spring.db')
            else:
                SpringDataBase().setup_sqlite_db(base)
                
        self.comm.barrier()
        session, ctf_parameters, micrograph_files = self.setup_database_and_ctfinfo(micrograph_files)
        
        micrograph_files = self.insure_that_every_node_has_the_same_number_of_micrographs(micrograph_files)
        
        self.log.plog(10)
        for each_micrograph_index, each_micrograph_file in enumerate(micrograph_files):
            if each_micrograph_file != 'place_holder':
                ctffind_parameters, ctftilt_parameters = \
                self.run_ctffind_and_ctftilt_for_each_micrograph(micrograph_files, outfiles, each_micrograph_index,
                each_micrograph_file)
            else:
                np_ctffind = self.make_ctffind_parameters_named_tuple()
                ctffind_parameters = np_ctffind._make(5 * [None])
                ctftilt_parameters = self.make_empty_ctftilt_parameters()
            
            self.comm.barrier()
            
            ctffinds = OpenMpi().convert_list_of_namedtuples_to_list_of_lists([ctffind_parameters])
            ctftilts = OpenMpi().convert_list_of_namedtuples_to_list_of_lists([ctftilt_parameters])
            
            those_micrographs = self.comm.gather(each_micrograph_file, root=0)
            ctffinds = self.comm.gather(ctffinds, root=0)
            ctftilts = self.comm.gather(ctftilts, root=0)
            if self.rank == 0:
                ctffinds = OpenMpi().merge_sequence_of_sequences(ctffinds)
                ctftilts = OpenMpi().merge_sequence_of_sequences(ctftilts)
                ctffind_tuple = self.make_ctffind_parameters_named_tuple()
                
                ctffind_params_combined_list = OpenMpi().convert_list_of_lists_to_list_of_provided_namedtuple(ctffinds,
                ctffind_tuple)
                
                ctftilt_tuple = self.make_ctftilt_parameters_named_tuple()
                
                ctftilt_params_combined_list = OpenMpi().convert_list_of_lists_to_list_of_provided_namedtuple(ctftilts,
                ctftilt_tuple)
            
                zipped_info = zip(those_micrographs, ctffind_params_combined_list, ctftilt_params_combined_list)
                for that_micrograph_file, each_ctffind, each_ctftilt in zipped_info:
                    if each_ctffind.defocus1 is not None:
                        session = self.enter_ctffind_values_in_database(session, that_micrograph_file,
                        self.ori_pixelsize, ctf_parameters, each_ctffind)
            
                    if each_ctftilt.defocus1 is not None:
                        session = self.enter_ctftilt_values_in_database(session, that_micrograph_file,
                        ctf_parameters.pixelsize, each_ctftilt)
            
                session.commit()
        
        self.ctftilt_parameters = ctftilt_parameters
        

    def determine_ctf(self):
        self.startup_scan_mpi_programs()
        
        if self.micrograph_files != []:
            self.run_ctffind_and_ctftilt_for_given_micrographs(self.micrograph_files, self.outfiles)
        
        self.comm.barrier()
                
        self.end_scan_mpi_programs()

def main():
    parset = MicCtfDeterminePar()
    reduced_parset = OpenMpi().start_main_mpi(parset)
    
    ####### Program
    micrograph = MicCtfDetermineMpi(reduced_parset)
    micrograph.determine_ctf()


if __name__ == '__main__':
    main()



# Author: Carsten Sachse 
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details
"""
Program to classify excised in-plane rotated segments using SPARX's k-means clustering
"""
from EMAN2 import Analyzers, EMData, EMUtil, Util
from collections import OrderedDict, namedtuple
from fundamentals import rot_shift2D
import os
import shutil
from spring.csinfrastr.csdatabase import SpringDataBase, base, SegmentTable
from spring.csinfrastr.csfeatures import Features
from spring.csinfrastr.cslogger import Logger
from spring.csinfrastr.csproductivity import OpenMpi, Support
from spring.csinfrastr.csreadinput import OptHandler
from spring.segment2d.segment import Segment
from spring.segment2d.segmentalign2d import SegmentAlign2dPar
from spring.segment2d.segmentexam import SegmentExam
from utilities import model_blank

from tabulate import tabulate

import numpy as np


class SegmentClassPar(SegmentAlign2dPar):
    """
    Class to initiate default dictionary with input parameters including help and range values and 
    status dictionary
    """
    def __init__(self):
        # package/program identity
        self.package = 'emspring'
        self.progname = 'segmentclass'
        self.proginfo = __doc__
        self.code_files = [self.progname]

        self.class_features = Features()
        self.feature_set = self.class_features.setup(self)
        
        self.define_parameters_and_their_properties()
        self.define_program_states()
    

    def add_mask_dimension_to_features(self):
        self.feature_set = self.class_features.set_helix_width_and_height(self.feature_set)
        

    def define_parameters_and_their_properties(self):
        self.feature_set = self.class_features.set_inp_stack(self.feature_set)
        self.feature_set = self.set_class_avg_out_stack(self.feature_set)
        
        self.feature_set = self.class_features.set_reference_option(self.feature_set)
        self.feature_set.relatives['Image reference stack'] = 'Reference image option'
        self.feature_set = self.class_features.set_image_reference_stack(self.feature_set, 'intermediate')
        self.feature_set = self.class_features.set_spring_db_option(self.feature_set, 'intermediate', False)
        self.feature_set = self.set_spring_db_prep_rec(self.feature_set)
        self.feature_set = self.class_features.set_spring_path(self.feature_set)

        self.feature_set = self.set_class_var_out_stack(self.feature_set)
        self.feature_set = self.set_eigenimg_out_stack(self.feature_set)
        self.feature_set = self.class_features.set_pixelsize(self.feature_set)
        
        self.add_mask_dimension_to_features()

        self.feature_set = self.set_class_count(self.feature_set)
        self.feature_set = self.set_iteration_count(self.feature_set)
        self.feature_set = self.set_keep_intermediate_files_option(self.feature_set)
        
        self.feature_set = self.class_features.set_alignment_rotation_and_translation(self.feature_set)
        
        self.feature_set = self.class_features.set_filter_options(self.feature_set)
        self.feature_set = self.set_frc_based_filter_option(self.feature_set)
        self.feature_set = self.class_features.set_binning_option(self.feature_set, default=True)
        self.feature_set = self.class_features.set_binning_factor(self.feature_set, binfactor=6, image='segments')
        self.feature_set = self.class_features.set_mpi(self.feature_set)
        self.feature_set = self.class_features.set_ncpus(self.feature_set)
        self.feature_set = self.class_features.set_temppath(self.feature_set)


    def define_program_states(self):
        self.feature_set.program_states['classify']='Classification of segments'
        self.feature_set.program_states['sx_kmeans']='K-means classification of segments'
        self.feature_set.program_states['spring_align2d']='Multi-reference alignment against obtained classes'
        self.feature_set.program_states['compute_eigenimages_from_aligned_stack']='Compute Eigenimages from aligned ' +\
        'stack'


    def set_class_avg_out_stack(self, feature_set):
        inp1 = 'Class average stack'
        feature_set.parameters[inp1] = 'averages.hdf'
        feature_set.properties[inp1] = feature_set.file_properties(1, ['spi', 'hdf', 'img', 'hed'], 'saveFile')
        
        feature_set.hints[inp1] = 'Stack: accepted image file formats ({0})'.\
        format(', '.join(feature_set.properties[inp1].ext))
        
        feature_set.level[inp1]='beginner'
        
        return feature_set


    def set_class_var_out_stack(self, feature_set):
        inp2 = 'Class variance stack'
        feature_set.parameters[inp2] = 'variances.hdf'
        feature_set.properties[inp2] = feature_set.file_properties(1, ['spi', 'hdf', 'img', 'hed'], 'saveFile')
        feature_set.hints[inp2] = 'Stack: accepted image file formats ({0})'.\
        format(', '.join(feature_set.properties[inp2].ext))
        
        feature_set.level[inp2]='expert'
        
        return feature_set


    def set_eigenimg_out_stack(self, feature_set):
        inp3 = 'Eigenimage stack'
        feature_set.parameters[inp3] = 'eigenimages.hdf'
        feature_set.properties[inp3] = feature_set.file_properties(1, ['spi', 'hdf', 'img', 'hed'], 'saveFile')
        feature_set.hints[inp3] = 'Stack: accepted image file formats ({0})'.\
        format(', '.join(feature_set.properties[inp3].ext))
        
        feature_set.level[inp3]='expert'
        
        return feature_set


    def set_class_count(self, feature_set):
        inp7 = 'Number of classes'
        feature_set.parameters[inp7] = int(5)
        feature_set.hints[inp7] = 'Number of classes to be clustered'
        feature_set.properties[inp7] = feature_set.Range(0, 2000, 1)
        feature_set.level[inp7]='beginner'
        
        return feature_set
    
    
    def set_iteration_count(self, feature_set):
        inp7 = 'Number of iterations'
        feature_set.parameters[inp7] = int(5)
        feature_set.hints[inp7] = 'Number of iteration cycles of clustering and alignment: 0 - only k-means ' + \
        'clustering, 1 - cluster and align, 2 - cluster/align/cluster/align... .'
        feature_set.properties[inp7] = feature_set.Range(0, 200, 1)
        feature_set.level[inp7]='intermediate'
        
        return feature_set
    

    def set_keep_intermediate_files_option(self, feature_set):
        inp15 = 'Keep intermediate files'
        feature_set.parameters[inp15]=bool(False)
        feature_set.hints[inp15]='Keep intermediate images from alignment ' + \
        'which are iteratively generated - EM image stacks are deleted otherwise.'
        
        feature_set.level[inp15]='intermediate'
        
        return feature_set
    
    
    def set_spring_db_prep_rec(self, feature_set):
        inp6 = 'Database prepare option'
        feature_set.parameters[inp6] = bool(True)
        feature_set.hints[inp6] = 'If checked will prepare database entries for 3D refinement using convolved images.'
        feature_set.level[inp6]='expert'
        feature_set.relatives[inp6]='Spring database option'
        
        return feature_set

    
class SegmentClassPreparation(object):
    """
    * Class that holds functions for examining segments from micrographs

    * __init__ Function to interpret multi-input parameters

    """

    def __init__(self, parset = None):
        self.log = Logger()
        if parset is not None:
            self.feature_set = parset
            self.p = self.feature_set.parameters

            self.infilestack = self.p['Image input stack']
            self.infile = self.p['Image input stack']
            self.output_avgstack = self.p['Class average stack']
            self.spring_db_option = self.p['Spring database option']
            self.spring_db_prepare = self.p['Database prepare option']
            self.spring_path = self.p['spring.db file']
            
            self.reference_option = self.p['Reference image option']
            self.reference_stack = self.p['Image reference stack']
            self.output_varstack = self.p['Class variance stack']
            self.eigenstack = self.p['Eigenimage stack']
#            self.center_option = self.p['Center option']
            self.pixelsize_ori = self.p['Pixel size in Angstrom']
            self.noclasses = self.p['Number of classes']
            self.iteration_count = self.p['Number of iterations']
            self.keep_intermediate_files = self.p['Keep intermediate files']
            
            self.restrain_inplane_rotation = self.p['Limit in-plane rotation']
            self.delta_psi = self.p['Delta in-plane rotation angle']
            self.x_range_in_angstrom = self.p['X and Y translation range in Angstrom'][0]
            self.y_range_in_angstrom = self.p['X and Y translation range in Angstrom'][1]
            
            self.setup_segmentalign_filter()
            self.mpi_option = self.p['MPI option']
            self.ncpus = self.p['Number of CPUs']
            self.temppath = self.p['Temporary directory']

            self.binoption=self.p['Binning option']
            self.binfactor=self.p['Binning factor']
            if self.binfactor == 1 and self.binoption is True:
                self.binoption = False

            self.define_helix_or_particle_dimensions()


    def setup_segmentalign_filter(self):
        self.high_pass_filter_option = self.p['High-pass filter option']
        self.low_pass_filter_option = self.p['Low-pass filter option']
        self.high_pass_filter_cutoff = self.p['High and low-pass filter cutoffs in 1/Angstrom'][0]
        self.low_pass_filter_cutoff = self.p['High and low-pass filter cutoffs in 1/Angstrom'][1]
        self.custom_filter_option = self.p['Custom filter option']
        self.custom_filter_file = self.p['Custom-built filter file']
        self.automatic_filter_option = self.p['Automatic filter option']
        self.bfactor = self.p['B-Factor']
        

    def define_helix_or_particle_dimensions(self):
        self.helixwidth, self.helixheight = self.p['Estimated helix width and height in Angstrom']
        self.helixwidthpix_ori = int(round(self.helixwidth / self.pixelsize_ori))
        self.helixheightpix_ori = int(round(self.helixheight / self.pixelsize_ori))


    def apply_binfactor_if_required(self):
        if self.binfactor > 1 and self.binoption is True:
            self.infilestack, self.image_dimension, self.helixwidthpix, self.pixelsize = \
            SegmentExam().apply_binfactor(self.binfactor, self.infilestack, self.image_dimension_ori,
            self.helixwidthpix_ori, self.pixelsize_ori)
            
            self.helixheightpix = int(self.helixheightpix_ori / self.binfactor)
        else:
            self.helixwidthpix = self.helixwidthpix_ori
            self.helixheightpix = self.helixheightpix_ori
            self.pixelsize = self.pixelsize_ori
            self.image_dimension = self.image_dimension_ori
        self.maskwidthpix = self.helixwidthpix
        self.maskwidthpix_ori = self.helixwidthpix_ori


    def determine_minimal_segment_size(self):
        min_segsize = int(np.sqrt(self.helixheightpix ** 2 + self.helixwidthpix ** 2) + 0.5)
        min_segsize = Segment().determine_boxsize_closest_to_fast_values(min_segsize)

        return min_segsize


    def perform_binning_and_trimming_of_input_stack_if_required(self):
        stack = EMData()
        stack.read_image(self.infilestack, 0)
        self.image_dimension_ori = stack.get_xsize()
        self.apply_binfactor_if_required()
        
        min_segsize = self.determine_minimal_segment_size()
        image_count = EMUtil.get_image_count(self.infilestack)
        for each_image_index in list(range(image_count)):
            stack.read_image(self.infilestack, each_image_index)
            if min_segsize < self.image_dimension:
                stack = Util.window(stack, min_segsize, min_segsize, 1, 0, 0, 0)
            stack.write_image(os.path.basename(self.infilestack), each_image_index)
            
        self.infilestack = os.path.basename(self.infilestack)
        if min_segsize < self.image_dimension:
            self.image_dimension = min_segsize
            
        return self.infilestack, self.image_dimension
    

    def perform_binning_and_trimming_of_reference_if_required(self):
        if self.binoption:
            self.reference_stack, image_dimension, self.maskwidthpix, self.pixelsize = \
            SegmentExam().apply_binfactor(self.binfactor, self.reference_stack, self.image_dimension_ori, 
            self.maskwidthpix_ori, self.pixelsize_ori)
            
        img = EMData()
        img.read_image(self.infilestack)
        trimmed_img_size = img.get_xsize()
        
        ref_img = EMData()
        ref_img.read_image(self.reference_stack)
        ref_img_size = ref_img.get_xsize()
        
        if trimmed_img_size < ref_img_size:
            ref_count = EMUtil.get_image_count(self.reference_stack)
            for each_img in list(range(ref_count)):
                ref_img.read_image(self.reference_stack, each_img)
                ref_img = Util.window(ref_img, trimmed_img_size, trimmed_img_size, 1, 0, 0, 0)
                ref_img.write_image(os.path.basename(self.reference_stack), each_img)
        
            self.reference_stack = os.path.basename(self.reference_stack)
        
        return self.reference_stack
            

    def prepare_mask_ori(self):
        self.mask_ori = SegmentExam().make_smooth_rectangular_mask(self.helixwidthpix_ori, self.helixheightpix_ori,
        self.image_dimension_ori)
        
        return self.mask_ori


    def prepare_mask(self):
        if self.helixheightpix > self.image_dimension:
            self.log.wlog('Specified helix height of {0} Angstrom = {1} '.format(self.helixheight, self.helixheightpix) + \
            'pixels is larger than provided image dimension {0} pixels'.format(self.image_dimension))
            self.helixheightpix = self.image_dimension
            
        self.mask = SegmentExam().make_smooth_rectangular_mask(self.helixwidthpix, self.helixheightpix,
        self.image_dimension)
        
        return self.mask


    def setup_mask(self):
        self.mask = self.prepare_mask()
        self.maskfile = 'rectmask.hdf'
        self.mask.write_image(self.maskfile)

        return self.maskfile


class SegmentClassExternalPrograms(object):

    def add_mask_dimensions_to_align2d_features(self):
        self.aligndict['Estimated helix width and height in Angstrom'] = ((self.helixwidth, self.helixheight))
        program_to_be_launched = 'segmentalign2d'
        
        return program_to_be_launched
        

    def run_segmentalign2d_in_separate_process(self, aligndict, align_directory, program_to_be_launched):
        cpu_count = aligndict['Number of CPUs']
        mpi_option = aligndict['MPI option']
        alistack = aligndict['Image output stack']

        segmentalign_parfile = Features().write_parfile(aligndict)
        command_line_string = ' --f {parfile} --d {directory}'.format(parfile=segmentalign_parfile,
        directory=align_directory)

        external_segmentalign_run = OpenMpi()
        if mpi_option is True:
            program_to_be_launched = program_to_be_launched + '_mpi'
            command_line_string = program_to_be_launched + command_line_string
            external_segmentalign_run.check_if_mpi_works_and_launch_command(command_line_string, cpu_count)
        else:
            command_line_string = program_to_be_launched + command_line_string
            external_segmentalign_run.launch_command(command_line_string)
        external_segmentalign_run.check_expected_output_file(program_to_be_launched, os.path.join(align_directory,
        alistack))

        alistack = os.path.join(align_directory, alistack)
        os.remove(segmentalign_parfile)

        return alistack


    def spring_align2d(self, infilestack, reference_stack, reference_option=True, local_refinement=False):
        """
        * Function to launch helical 2D alignment program segmentalign

        #. Input: infilestack, reference stack
        #. Output: aligned stack
        #. Usage: alistack = spring_align2d(infilestack, reference_stack)
        """

        self.log.fcttolog()
        
        self.aligndict = OrderedDict()
        self.aligndict['Image input stack'] = infilestack
        alistack = os.path.splitext(os.path.basename(infilestack))[0] + 'ali.hdf'
        self.aligndict['Image output stack'] = alistack
        self.aligndict['Number of iterations']= 3
        self.aligndict['Internal binning factor']= 1
        self.aligndict['Update references']=True
        self.aligndict['Aligned average stack'] = 'multi_ref_avg.hdf'
        self.aligndict['Image reference stack'] = reference_stack
        self.aligndict['Reference option'] = reference_option 
        self.aligndict['Pixel size in Angstrom'] = self.pixelsize
        
        program_to_be_launched = self.add_mask_dimensions_to_align2d_features()
        
        self.aligndict['Limit in-plane rotation']=self.restrain_inplane_rotation
        self.aligndict['Delta in-plane rotation angle'] = self.delta_psi
        self.aligndict['X and Y translation range in Angstrom'] = ((self.x_range_in_angstrom, self.y_range_in_angstrom))
    
        self.aligndict['Local refinement'] = local_refinement
        self.aligndict['Absolute X and Y translation limit in Angstrom']= ((self.x_range_in_angstrom, self.y_range_in_angstrom))
        
        self.aligndict['High-pass filter option']=self.high_pass_filter_option
        self.aligndict['Low-pass filter option']=self.low_pass_filter_option
        self.aligndict['High and low-pass filter cutoffs in 1/Angstrom']=((self.high_pass_filter_cutoff, 
                                                                           self.low_pass_filter_cutoff))
        self.aligndict['Custom filter option']=self.custom_filter_option
        self.aligndict['Custom-built filter file']=self.custom_filter_file
        self.aligndict['Automatic filter option']=self.automatic_filter_option
        self.aligndict['B-Factor']=self.bfactor
        
        self.aligndict['MPI option'] = self.mpi_option
        self.aligndict['Number of CPUs']=self.ncpus
        self.aligndict['Temporary directory'] = self.temppath 
        
#         segmentalign_parfile = Features().write_parfile(self.aligndict)
        self.align_directory = '{0}{1:02}'.format(program_to_be_launched, self.iteration_index)
#         abspath_logfile = os.path.join(os.path.split(os.path.abspath(os.curdir))[0], self.feature_set.logfile)
        
#         command_line_string = ' --f {parfile} --d {directory}'.format(parfile=segmentalign_parfile,
#         directory=self.align_directory, logfile=abspath_logfile)

        alistack = self.run_segmentalign2d_in_separate_process(self.aligndict, self.align_directory,
        program_to_be_launched)

        return alistack
    

    def divide_sx_kmeans_in_reasonable_chunks(self, img_count, class_count, chunk_size=2000):
        """
        >>> from spring.segment2d.segmentclass import SegmentClass
        >>> SegmentClass().divide_sx_kmeans_in_reasonable_chunks(10, 3, 6)
        ([[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]], [3])
        >>> SegmentClass().divide_sx_kmeans_in_reasonable_chunks(10, 4, 6)
        ([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]], [2, 2])
        >>> SegmentClass().divide_sx_kmeans_in_reasonable_chunks(20, 6, 5)
        ([[0, 1, 2, 3, 4, 5, 6], [7, 8, 9, 10, 11, 12], [13, 14, 15, 16, 17, 18, 19]], [2, 2, 2])
        """
        parts_for_proc = img_count / chunk_size + 1
        if class_count / parts_for_proc < 2:
            parts_for_proc = class_count / 2
        
        images = list(range(img_count))
        classes = list(range(class_count))
        
        split_imgs = OpenMpi().split_sequence_evenly(images, parts_for_proc)
        split_classes = OpenMpi().split_sequence_evenly(classes, parts_for_proc)
        
        class_counts = [len(each_split_class) for each_split_class in split_classes]
        
        return split_imgs, class_counts
        
        
    def sx_kmeans_wrap(self, alistack, maskfile, class_count):
        img_count = EMUtil.get_image_count(alistack)
        
        if hasattr(self, 'test_cls_count'):
            split_imgs, class_counts = self.divide_sx_kmeans_in_reasonable_chunks(img_count, class_count, chunk_size=150)
        else: 
            split_imgs, class_counts = self.divide_sx_kmeans_in_reasonable_chunks(img_count, class_count)

        class_parts = len(class_counts)
        if class_parts > 1:
            classdir_common = self.get_dir_name_for_classification('sxk_means.py', self.iteration_index)
            workdir = os.path.abspath((os.curdir))
            alistack = os.path.join(workdir, alistack)
            maskfile = os.path.join(workdir, maskfile)
            os.mkdir(classdir_common)
            os.chdir(classdir_common)

            img = EMData()
            for each_run_id, (each_imgs, each_class_count) in enumerate(zip(split_imgs, class_counts)):
                local_ali_stack = os.path.splitext(alistack)[0] + '_{0:03}'.format(each_run_id) + os.path.splitext(alistack)[-1]
                for each_local_id, each_img_id in enumerate(each_imgs):
                    img.read_image(alistack, each_img_id)
                    img.write_image(local_ali_stack, each_local_id)
                
                local_avg_stack, local_var_stack = self.sx_kmeans(local_ali_stack, maskfile, each_class_count, each_run_id)
    
                comb_avg_stack = os.path.basename(local_avg_stack)
                comb_var_stack = os.path.basename(local_var_stack)

                if each_run_id == 0:
                    current_cls_count = 0
                else:
                    current_cls_count = EMUtil.get_image_count(comb_avg_stack)
    
                for each_class in list(range(each_class_count)):
                    img.read_image(local_avg_stack, each_class)
                    members = img.get_attr('members')
                    members_updated = [each_imgs[0] + each_member for each_member in members]
                    img.set_attr('members', members_updated)
                    img.set_attr('ave_n', len(members_updated))
                    img.write_image(comb_avg_stack, current_cls_count + each_class)
                    
                    img.read_image(local_var_stack, each_class)
                    img.write_image(comb_var_stack, current_cls_count + each_class)
    
                os.remove(local_ali_stack)
                os.remove(local_avg_stack)
                os.remove(local_var_stack)
                os.rmdir(os.path.dirname(local_avg_stack))

            comb_avg_stack = os.path.abspath(comb_avg_stack)
            comb_var_stack = os.path.abspath(comb_var_stack)
            os.chdir(workdir)

        if class_parts == 1:
            local_avg_stack, local_var_stack = self.sx_kmeans(alistack, maskfile, class_counts[0])
            comb_avg_stack = os.path.abspath(local_avg_stack)
            comb_var_stack = os.path.abspath(local_var_stack)

        return comb_avg_stack, comb_var_stack

                    
    def get_dir_name_for_classification(self, program_to_be_launched, iteration_id, run_id=None):
        if run_id is None:
            classdir = '{0}{1:02}'.format(os.path.splitext(program_to_be_launched)[0], iteration_id)
        else:
            classdir = '{0}{1:02}_{2:03}'.format(os.path.splitext(program_to_be_launched)[0], iteration_id, run_id)

        return classdir


    def sx_kmeans(self, alistack, maskfile, class_count, run_id=None):
        """
        * Function to launch SPARX's k-means analysis

        #. Input: aligned stack, mask file, number of classes 
        #. Output: stack of averages, stack of variances
        #. Usage: avgstack, varstack = sx_kmeans(alistack, maskfile, class_count)
        """

        self.log.fcttolog()
        external_kmeans_run = OpenMpi()
        
        program_to_be_launched = 'sxk_means.py'
        classdir = self.get_dir_name_for_classification(program_to_be_launched, self.iteration_index, run_id)
        program_to_be_launched = Support().search_path_like_which(program_to_be_launched)
        avgstack = os.path.join(classdir, 'averages.hdf')
        varstack = os.path.join(classdir, 'variances.hdf')
        command_line_string = 'springenv {0} {1} {2} '.format(program_to_be_launched, alistack, classdir) + \
        '{0} --K={1} --maxit=500 --rand_seed=-1 --crit=D'.format(maskfile, class_count)

        if self.mpi_option:
            command_line_string = command_line_string + ' --MPI'
            external_kmeans_run.check_if_mpi_works_and_launch_command(command_line_string, self.ncpus)
        else:
            external_kmeans_run.launch_command(command_line_string)

        self.log.tlog(command_line_string)

        external_kmeans_run.check_expected_output_file(program_to_be_launched, avgstack)

        return avgstack, varstack
    
  
    def sx_kmeans_groups(self, alistack=None, maskfile=None, noclasses=None):
        """
        * Function to launch SPARX's k-means groups analysis (sx_kmeans_groups.py)

        #. Input: aligned stack, mask file, number of classes 
        #. Output: statistics data
        #. Usage: avgstack, varstack = sx_kmeans_groups(alistack, maskfile, noclasses)
        """
        if alistack is None: alistack = self.alistack
        if maskfile is None: maskfile = self.maskfile
        if noclasses is None: noclasses = self.noclasses

        self.classgrpdir = 'sxk_means_groups'
        external_kmeans_group_run = OpenMpi()
        self.classgrpdir = OpenMpi().check_dir(self.classgrpdir)

        program_to_be_launched = Support().search_path_like_which('sxk_means_groups.py')
        command_line_string = 'springenv {program} {instack} {outdir} {maskfile} --K1=2 \
            --K2={noclasses} --maxit=100 --rand_seed=100 \
            '.format(instack=alistack, program=program_to_be_launched, 
            outdir=self.classgrpdir, maskfile=maskfile, noclasses=noclasses)

        self.log.tlog(command_line_string)
        if self.mpi_option is True:
            command_line_string = command_line_string + ' --MPI'
            external_kmeans_group_run.check_if_mpi_works_and_launch_command(command_line_string, self.ncpus)
        else:
            external_kmeans_group_run.launch_command(command_line_string)


class SegmentClassStatistics(object):
    def compute_eigen_images(self, stack, eigenstack, mask, avg):
        """
          Perform PCA on stack file 
          and Get eigen images
        """
    
        image_count = EMUtil.get_image_count(stack)
        if image_count > 2000:
            eigenimage_count = 200
        else:
            eigenimage_count = int(0.1*image_count)
            
        pca = Analyzers.get('pca_large', {'mask':mask, 'nvec':eigenimage_count})
        e = EMData()
        if avg == 1: 
            s = EMData()
        for each_img_id in list(range(image_count)):
            e.read_image(stack, each_img_id)
            if mask is not None:
                e *= mask
            pca.insert_image(e)
            if avg == 1:
                if each_img_id == 0: 
                    s  = pca
                else:      
                    s += pca
        if avg == 1: 
            pca -= s/image_count
        eigenimages = pca.analyze()
        
        for each_img_id, each_eigenimage in enumerate(eigenimages):
            each_eigenimage.write_image(eigenstack,each_img_id)
        
        return eigenstack
    
        
    def log_member_statistics_from_classes(self, class_avg_stack, source_stack):
        
        class_avg_stack_image = EMData()
        class_count = EMUtil.get_image_count(class_avg_stack)
        raw_image_count = EMUtil.get_image_count(source_stack)
        
        members_of_classes = []
        classes =[]
        log_info = []
        for each_class in list(range(class_count)):
            class_avg_stack_image.read_image(class_avg_stack, each_class)
            members = class_avg_stack_image.get_attr('members')
            if type(members) is list:
                members = [int(each_member) for each_member in members]
                members_str = ', '.join([str(each_member) for each_member in members])
                
                log_info += [[each_class, len(members), len(members) * 100 / float(raw_image_count), members_str]]
                
                this_class = [each_class] * len(members)
                members_class = zip(members, this_class)
                members_of_classes += (members_class)
                classes.append(members)
            elif type(members) is not list:
                log_info += [[each_class, 0, 0, each_class, '0']]
        
        msg = tabulate(log_info, ['class_id', 'member_count', 'member_ratio', 'members'])
        self.log.ilog('The following classes and the corresponding membership statistics are listed:\n{0}'.format(msg))
        
        return members_of_classes, classes
    
            
    def enter_class_assignment_in_database(self, session, members_class):
        for each_member, each_class in members_class:
            each_segment = session.query(SegmentTable).get(each_member + 1)
            each_segment.class_id = each_class
            if self.spring_db_prepare:
                each_segment.ctf_convolved = True
                each_segment.ctf_phase_flipped = False
                each_segment.x_coordinate_A = each_segment.picked_x_coordinate_A
                each_segment.y_coordinate_A = each_segment.picked_y_coordinate_A
            session.merge(each_segment)
        session.commit()
            
        
    def make_align_named_tuple(self):
        return namedtuple('align', 'id alpha tx ty mirror')
    
    
    def get_alignment_info_from_stack(self, aligned_stack, classes):
        algn_img = EMData()
        alignment_data = []
        align_nt = self.make_align_named_tuple()
        for each_cls in classes:
            alignment_cls = []
            for each_member_id in each_cls:
                algn_img.read_image(aligned_stack, each_member_id)
                if algn_img.has_attr('xform.align2d'):
                    trans = algn_img.get_attr('xform.align2d')
                    d = trans.get_params('2D')
                    align_params = align_nt(each_member_id, d['alpha'], self.binfactor * d['tx'], 
                                            self.binfactor * d['ty'], d['mirror'])
                else:
                    align_params = align_nt(each_member_id, 0, 0, 0, 0)
                alignment_cls.append(align_params)
            
            alignment_data.append(alignment_cls)
        
        return alignment_data


    def compute_average_variance_and_eigenimages_on_orignal_stack(self, output_avgstack, output_varstack, classes,
    alignment_data):
        self.log.fcttolog()
        ori_img = EMData()
        ori_img.read_image(self.infile)
        
        aligned_stack = 'aligned_ori_stack.hdf'
        segment_size = ori_img.get_xsize()
        avg = model_blank(segment_size, segment_size, 1)
        for each_cls_id, each_cls in enumerate(alignment_data):
            cls_avg = model_blank(segment_size, segment_size, 1)
            cls_var = cls_avg.copy()
            for each_img in each_cls:
                ori_img.read_image(self.infile, each_img.id)
                img = rot_shift2D(ori_img, each_img.alpha, each_img.tx, each_img.ty, each_img.mirror)
                img.write_image(aligned_stack, each_img.id)
                
                avg += img
                cls_avg += img
                cls_var += img * img
            
            member_count = len(each_cls)
            cls_avg /= float(member_count)
            cls_avg.set_attr('members', classes[each_cls_id])
            cls_avg.set_attr('n_ave', len(classes[each_cls_id]))
            cls_avg.write_image(output_avgstack, each_cls_id)
            cls_var = (cls_var - cls_avg * cls_avg * member_count) / float(member_count - 1)
            cls_var.write_image(output_varstack, each_cls_id)
        self.log.tlog('Averages and variances of aligned stack computed.')

        self.mask_ori = self.prepare_mask_ori()
        self.compute_eigen_images(aligned_stack, self.eigenstack, self.mask_ori, avg)
        self.log.tlog('Eigenimages of aligned stack computed.')
        
        os.remove(aligned_stack)
        

    def create_database_with_stack_id_entries(self, segment_count):
        session = SpringDataBase().setup_sqlite_db(base)
        for each_stack_id in list(range(segment_count)):
            each_segment = SegmentTable()
            each_segment.stack_id = each_stack_id
            session.add(each_segment)
        
        session.commit()
        
        return session
    
    
    def finish_classification(self, avgstack, varstack, aligned_stack, output_avgstack, output_varstack):
        os.remove(self.maskfile)
        if self.iteration_count != 0:
            avg_file = os.path.join(self.align_directory, 'multi_ref_avg.hdf')
        else:
            avg_file = avgstack
        
        members_of_classes, classes = self.log_member_statistics_from_classes(avg_file, aligned_stack)
        
        alignment_data = self.get_alignment_info_from_stack(aligned_stack, classes)
                
        self.compute_average_variance_and_eigenimages_on_orignal_stack(output_avgstack, output_varstack, classes,
        alignment_data)
        
        if self.spring_db_option:
            shutil.copy(self.spring_path, 'spring.db')
            session = SpringDataBase().setup_sqlite_db(base)
        else:
            segment_count = EMUtil.get_image_count(self.infilestack)
            session = self.create_database_with_stack_id_entries(segment_count)
        self.enter_class_assignment_in_database(session, members_of_classes)
            

class SegmentClass(SegmentClassPreparation, SegmentClassExternalPrograms, SegmentClassStatistics):

    def classify(self):
        self.log.fcttolog()

        self.infilestack, self.image_dimension = self.perform_binning_and_trimming_of_input_stack_if_required()
        self.maskfile = self.setup_mask()

        self.log.plog(10)
        self.iteration_index = 0
        if self.reference_option:
            self.reference_stack = self.perform_binning_and_trimming_of_reference_if_required()
                
            aligned_stack = self.spring_align2d(self.infilestack, self.reference_stack, reference_option=True,
            local_refinement=False)
        else:
            aligned_stack = self.infilestack
            
        avgstack, varstack = self.sx_kmeans_wrap(aligned_stack, self.maskfile, self.noclasses)
        self.log.plog(30)
        
        for self.iteration_round in list(range(self.iteration_count)):
            self.iteration_index += 1
            if self.iteration_round != 0:
                avgstack, varstack = self.sx_kmeans_wrap(aligned_stack, self.maskfile, self.noclasses)
                if not self.keep_intermediate_files:
                    os.remove(aligned_stack)
                
            self.log.plog(80 * (self.iteration_round + 0.5) / self.iteration_count + 30)
            if self.iteration_round == 0 and not self.reference_option:
                aligned_stack = self.spring_align2d(self.infilestack, avgstack, reference_option=True,
                local_refinement=False)
            else:
                aligned_stack = self.spring_align2d(self.infilestack, avgstack, reference_option=True,
                local_refinement=True)
                
            self.log.plog(80 * (self.iteration_round + 1) / self.iteration_count + 30)
        
        self.finish_classification(avgstack, varstack, self.infilestack, self.output_avgstack, self.output_varstack)
        
        self.log.endlog(self.feature_set)


def main():
    # Option handling
    parset = SegmentClassPar()
    mergeparset = OptHandler(parset)

    ######## Program
    stack = SegmentClass(mergeparset)
    stack.classify()

if __name__ == '__main__':
    main()



# Author: Carsten Sachse 
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details
"""
Program to align segments from helical specimens with a restrained in-plane rotation of 0 or 180 +/- delta degrees
"""
from EMAN2 import EMData, EMUtil, Util
from alignment import Numrinit, ringwe
from collections import namedtuple, defaultdict
from fundamentals import rot_shift2D
from spring.csinfrastr.csfeatures import Features
from spring.csinfrastr.cslogger import Logger
from spring.csinfrastr.csproductivity import OpenMpi, Temporary
from spring.segment2d.segment import Segment
from spring.segment2d.segmentexam import SegmentExam
from tabulate import tabulate
from utilities import model_blank, compose_transform2, get_params2D
import numpy as np
import os


class SegmentAlign2dPar(object):
    """
    Class to initiate default dictionary with input parameters including help and range values and status dictionary
    """
    def __init__(self):
        # package/program identity
        self.package = 'emspring'
        self.progname = 'segmentalign2d'
        self.proginfo = __doc__
        self.code_files = ['segmentalign2d_prep', self.progname, self.progname + '_mpi']

        self.align2d_features = Features()
        self.feature_set = self.align2d_features.setup(self)
        
        self.define_parameters_and_their_properties()
        self.define_program_states()
    

    def add_mask_dimension_to_features(self):
        self.feature_set = self.align2d_features.set_helix_width_and_height(self.feature_set)
        

    def define_parameters_and_their_properties(self):
        self.feature_set = self.align2d_features.set_inp_stack(self.feature_set)
        self.feature_set = self.align2d_features.set_out_stack(self.feature_set)
        self.feature_set = self.set_iteration_count(self.feature_set)
        
        self.feature_set = self.set_reference_option(self.feature_set)

        self.feature_set = self.align2d_features.set_image_reference_stack(self.feature_set, 'beginner', 
        'Reference option')

        self.feature_set = self.align2d_features.set_reference_update_option(self.feature_set)
        self.feature_set = self.align2d_features.set_reference_output_aligned(self.feature_set)
        self.feature_set = self.align2d_features.set_pixelsize(self.feature_set)
        
        self.add_mask_dimension_to_features()

        self.feature_set = self.set_refinement_option(self.feature_set)
        self.feature_set = self.align2d_features.set_absolute_translation_limit(self.feature_set, 'Local refinement')
        self.feature_set = self.align2d_features.set_internal_binning(self.feature_set)
        self.feature_set = self.align2d_features.set_alignment_rotation_and_translation(self.feature_set)
        
        self.feature_set = self.align2d_features.set_filter_options(self.feature_set)
        self.feature_set = self.set_frc_based_filter_option(self.feature_set)
        
        self.feature_set = self.align2d_features.set_mpi(self.feature_set)
        self.feature_set = self.align2d_features.set_ncpus(self.feature_set)
        self.feature_set = self.align2d_features.set_temppath(self.feature_set)
        

    def set_reference_option(self, feature_set):
        inp14 = 'Reference option'
        feature_set.parameters[inp14] = True
        
        feature_set.hints[inp14] = 'Provide a stack of images as a reference. If not ticked, a simple average of ' +\
        'the provided image stack will be used.'
        
        feature_set.level[inp14]='beginner'
        
        return feature_set
    
        
    def set_iteration_count(self, feature_set):
        inp7 = 'Number of iterations'
        feature_set.parameters[inp7] = int(2)
        feature_set.hints[inp7] = 'Number of iteration cycles of multi-reference alignment. Internally images' + \
        'will be binned to speed up processing.'
        feature_set.properties[inp7] = feature_set.Range(1, 200, 1)
        feature_set.level[inp7]='intermediate'
        
        return feature_set
    
    
    def set_refinement_option(self, feature_set):
        inp14 = 'Local refinement'
        feature_set.parameters[inp14] = False
        
        feature_set.hints[inp14] = 'This option can be used to refine locally around the alignment parameters ' + \
        'stored in the header of the image stack.'
        
        feature_set.level[inp14]='expert'
        
        return feature_set
    
    
    def set_frc_based_filter_option(self, feature_set):
        inp14 = 'Automatic filter option'
        feature_set.parameters[inp14] = bool(True)
        
        feature_set.hints[inp14] = 'Automatic filter design derived from decay of Fourier ring correlation between ' + \
        'images.'
        
        feature_set.level[inp14]='expert'
        
        return feature_set
    

    def define_program_states(self):
        self.feature_set.program_states['prepare_alignment']='Prepare alignment'
        self.feature_set.program_states['prepare_reference_images_for_alignment']='Prepare reference images ' + \
        'for alignment'
        
        self.feature_set.program_states['align_images_to_references']='Aligns images to reference'
        
        self.feature_set.program_states['pass_alignment_parameters_from_reference_groups_to_images']='Apply ' + \
        'alignment parameters to images'

class SegmentAlign2dPreparation(object):
    """
    * Class to prepare iterative alignment of segments
    """

    def __init__(self, parset = None):
        self.log = Logger()
        if parset is not None:
            self.feature_set = parset
            self.p = self.feature_set.parameters

            self.set_given_parameters()
#            self.set_hidden_parameters()
            

    def define_helix_or_particle_dimensions(self):
        self.helixwidth, self.helixheight = self.p['Estimated helix width and height in Angstrom']
        self.helixwidthpix = int(round(self.helixwidth / self.pixelsize))
        self.helixheightpix = int(round(self.helixheight / self.pixelsize))

    def set_given_parameters(self):
        self.alignment_stack_name = self.p['Image input stack']
        self.infile = self.alignment_stack_name
        self.outfile = self.p['Image output stack']
        self.number_of_iterations=self.p['Number of iterations']
        
        self.reference_stack_name = self.p['Image reference stack']
        self.update_references = self.p['Update references']
        self.aligned_averages = self.p['Aligned average stack']
        self.pixelsize = self.p['Pixel size in Angstrom']
        self.reference_option = self.p['Reference option']
        self.define_helix_or_particle_dimensions()
        
        self.binfactor = self.p['Internal binning factor']
        
        self.restrain_inplane_rotation = self.p['Limit in-plane rotation']
        self.delta_psi = self.p['Delta in-plane rotation angle']
        if not self.restrain_inplane_rotation:
            self.delta_psi = 180
        self.x_range_A = self.p['X and Y translation range in Angstrom'][0]
        self.y_range_A = self.p['X and Y translation range in Angstrom'][1]
        
        self.refine_locally = self.p['Local refinement']
        self.x_limit_A, self.y_limit_A = self.p['Absolute X and Y translation limit in Angstrom']
        if not self.refine_locally:
            self.x_limit_A = self.x_range_A
            self.y_limit_A = self.y_range_A
        
        self.low_pass_filter_option = self.p['Low-pass filter option']
        self.high_pass_filter_option = self.p['High-pass filter option']
        self.high_pass_filter_cutoff = self.p['High and low-pass filter cutoffs in 1/Angstrom'][0]
        self.low_pass_filter_cutoff = self.p['High and low-pass filter cutoffs in 1/Angstrom'][1]
        self.custom_filter_option = self.p['Custom filter option']
        self.custom_filter_file = self.p['Custom-built filter file']
        self.bfactor = self.p['B-Factor']
        self.frc_filter_option = self.p['Automatic filter option']
        self.mpi_option = self.p['MPI option']
        self.cpu_count = self.p['Number of CPUs']
        self.temppath = self.p['Temporary directory']
        
        
    def prepare_cosine_falloff(self, image_dimension, start_falloff, falloff_length):
        """
        >>> from spring.segment2d.segmentalign2d import SegmentAlign2d
        >>> SegmentAlign2d().prepare_cosine_falloff(30, 7, 5)
        array([ 1.       ,  1.       ,  1.       ,  1.       ,  1.       ,
                1.       ,  1.       ,  0.9330127,  0.75     ,  0.5      ,
                0.25     ,  0.0669873,  0.       ,  0.       ,  0.       ])
        >>> SegmentAlign2d().prepare_cosine_falloff(50, 21, 10)
        array([ 1.        ,  1.        ,  1.        ,  1.        ,  1.        ,
                1.        ,  1.        ,  1.        ,  1.        ,  1.        ,
                1.        ,  1.        ,  1.        ,  1.        ,  1.        ,
                1.        ,  1.        ,  1.        ,  1.        ,  1.        ,
                1.        ,  0.97974649,  0.92062677,  0.82743037,  0.70770751])

        """
        coeff = np.ones(image_dimension / 2)
        coeff[start_falloff:] = 0.0
        
        falloff = (np.cos(np.linspace(0, 1, falloff_length + 2) * np.pi) + 1) / 2
        
        end_falloff = min(start_falloff + falloff_length, len(coeff))
        diff_length = len(coeff) - (start_falloff + falloff_length)

        if diff_length < 0:
            premature_end = falloff_length + diff_length + 1
            true_falloff = falloff[1:premature_end]
        else: 
            true_falloff = falloff[1:-1]
            
        coeff[start_falloff:end_falloff]=true_falloff
        
        return coeff
    
        
    def prepare_bfactor_coefficients(self, bfactor, pixelsize, image_dimension, cutoff_A=None):
        """
        >>> from spring.segment2d.segmentalign2d import SegmentAlign2d
        >>> SegmentAlign2d().prepare_bfactor_coefficients(0, 1.0, 50)
        array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,
                1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.])
        >>> SegmentAlign2d().prepare_bfactor_coefficients(100, 1.0, 50)
        array([ 1.        ,  0.98920796,  0.95752564,  0.90696062,  0.84062374,
                0.76241263,  0.67663385,  0.58761458,  0.49935179,  0.41523683,
                0.33787832,  0.26902956,  0.20961139,  0.15981037,  0.1192258 ,
                0.08703837,  0.06217652,  0.04346276,  0.02972922,  0.0198987 ,
                0.01303291,  0.00835282,  0.00523842,  0.00321471,  0.00193045])
        >>> SegmentAlign2d().prepare_bfactor_coefficients(-100, 1.0, 50)
        array([   1.        ,    1.01090978,    1.04435845,    1.10258371,
                  1.18959286,    1.3116257 ,    1.4779042 ,    1.70179577,
                  2.00259621,    2.40826423,    2.95964534,    3.7170637 ,
                  4.77073318,    6.25741626,    8.38744639,   11.48918606,
                 16.08324067,   23.00820047,   33.63694445,   50.25453012,
                 76.72884995,  119.72006788,  190.89740003,  311.06981145,
                518.01282467])
        >>> SegmentAlign2d().prepare_bfactor_coefficients(0, 1.0, 50, 4.0)
        array([ 1.       ,  1.       ,  1.       ,  1.       ,  1.       ,
                1.       ,  1.       ,  1.       ,  1.       ,  1.       ,
                1.       ,  1.       ,  1.       ,  0.9330127,  0.75     ,
                0.5      ,  0.25     ,  0.0669873,  0.       ,  0.       ,
                0.       ,  0.       ,  0.       ,  0.       ,  0.       ])
        >>> SegmentAlign2d().prepare_bfactor_coefficients(100, 1.0, 50, 4.0)
        array([ 1.        ,  0.98920796,  0.95752564,  0.90696062,  0.84062374,
                0.76241263,  0.67663385,  0.58761458,  0.49935179,  0.41523683,
                0.33787832,  0.26902956,  0.20961139,  0.1491051 ,  0.08941935,
                0.04351918,  0.01554413,  0.00291145,  0.        ,  0.        ,
                0.        ,  0.        ,  0.        ,  0.        ,  0.        ])
        >>> SegmentAlign2d().prepare_bfactor_coefficients(0.0, 5.0, 50, 12)
        array([ 1.       ,  1.       ,  1.       ,  1.       ,  1.       ,
                1.       ,  1.       ,  1.       ,  1.       ,  1.       ,
                1.       ,  1.       ,  1.       ,  1.       ,  1.       ,
                1.       ,  1.       ,  1.       ,  1.       ,  1.       ,
                1.       ,  0.9330127,  0.75     ,  0.5      ,  0.25     ])

        """
        """
        Niko Grigorieff's bfactor.exe::
        
            % bfactor.exe
            Input 3D map?
            circle.spi
              OPENING SPIDER FORMAT FILE....
             B-factor in A^2
            100
             Gauss filter (1) or cosine edge mask (2)?
            
            2
             Cosine edge filter radius in A?
            1
             Width of edge in pixels?
            1
             Pixel size in A?
            1
             Output 3D map?
            new.spi
              OPENING SPIDER FORMAT FILE....
                
                ===========  ================ 
                 Resolution  Weighting Factor
                ===========  ================ 
                      50.00            1.0000
                      25.00            0.9900
                      16.67            0.9608
                      12.50            0.9139
                      10.00            0.8521
                       8.33            0.7788
                       7.14            0.6977
                       6.25            0.6126
                       5.56            0.5273
                       5.00            0.4449
                       4.55            0.3679
                       4.17            0.2982
                       3.85            0.2369
                       3.57            0.1845
                       3.33            0.1409
                       3.12            0.1054
                       2.94            0.0773
                       2.78            0.0556
                       2.63            0.0392
                       2.50            0.0271
                       2.38            0.0183
                       2.27            0.0122
                       2.17            0.0079
                       2.08            0.0050
                       2.00            0.0032
                       1.92            0.0019
                ===========  ================
        
        Niko Grigorieff's bfactor.exe::
        
            % bfactor.exe
             Input 3D map?
            circle.spi
              OPENING SPIDER FORMAT FILE....
             B-factor in A^2
            -100
             Gauss filter (1) or cosine edge mask (2)?
            1
             Gauss filter radius in A?
            1
             Pixel size in A?
            1
             Output 3D map?
            new.spi
              OPENING SPIDER FORMAT FILE....
                
                ===========  ================
                 Resolution  Weighting Factor
                ===========  ================
                      50.00            1.0000
                      25.00            1.0098
                      16.67            1.0400
                      12.50            1.0922
                      10.00            1.1698
                       8.33            1.2776
                       7.14            1.4230
                       6.25            1.6164
                       5.56            1.8724
                       5.00            2.2118
                       4.55            2.6645
                       4.17            3.2733
                       3.85            4.1009
                       3.57            5.2394
                       3.33            6.8264
                       3.12            9.0703
                       2.94           12.2902
                       2.78           16.9828
                       2.63           23.9316
                       2.50           34.3912
                       2.38           50.4004
                       2.27           75.3241
                       2.17          114.8010
                       2.08          178.4306
                       2.00          282.8170
                       1.92          457.1447
                ===========  ================
        """
        one_over_res = SegmentExam().make_oneoverres(np.arange(image_dimension / 2), pixelsize)
        fourier_coefficients = np.exp((-bfactor / 4.0) * one_over_res**2)
        
        if cutoff_A > 2 * pixelsize:
            start_falloff = int(round(image_dimension * pixelsize / float(cutoff_A)))
            falloff_coeff = self.prepare_cosine_falloff(image_dimension, start_falloff, 5)
            fourier_coefficients *= falloff_coeff
        
        return fourier_coefficients
    
    
    def prepare_filter_function(self, high_pass_filter_option, high_pass_filter_cutoff, low_pass_filter_option,
    low_pass_filter_cutoff, pixelsize, image_dimension, filter_falloff = 0.08, custom_filter_option=False,
    custom_filter_file=None, bfactor=0.0):
        """
        * Function to generate a filter function based on hyperbolic tangent \
        (low-pass, high-pass or band-pass filters are possible)
        
        #. Input: high_pass_filter_option, low_pass_filter_option, 
        custom_filter_option: True or False, high_pass_filter_cutoff, 
        low_pass_filter_cutoff: in 1/Angstrom, pixelsize: Angstrom/pixel, 
        image_dimension: pixel, pixelsize, image_dimension: number of pixels
        filter_falloff: percent of pixels that make up smooth filter falloff
        custom_filter_file: recognizes last column as filter coefficient from 
        provided text file
                    
        #. Output: list of filter coefficients as in Fourier pixels
        
        >>> from spring.segment2d.segmentalign2d import SegmentAlign2d
        >>> sa = SegmentAlign2d()
        >>> sa.prepare_filter_function(True, 0.02, True, 0.08, 5, 10, 0.08) #doctest: +NORMALIZE_WHITESPACE
        [8.2287160299010509e-06, 0.99999853615457568, 1.0, 0.99999853615457568, 
        4.1143580149505254e-06]

        >>> sa.prepare_filter_function(True, 0.04, True, 0.06, 5, 50, 0.05) #doctest: +NORMALIZE_WHITESPACE
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 2.4444668511591772e-11, 
        0.00092821640674234374, 0.99997168174280127, 0.99999999999925493, 1.0, 
        1.0, 1.0, 0.99999999999925493, 0.99997168174280127, 0.00092821640674228822, 
        2.4444668511591772e-11, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        >>> sa.prepare_filter_function(True, 0.04, False, 0.06, 1.5, 50, 0.05, False, None, -100) #doctest: +NORMALIZE_WHITESPACE
        [5.172734356748137e-05, 0.002565400437652817, 0.051660149853613654, 
        0.06479161976119088, 0.06716260031683237, 0.07014331907922618, 
        0.07396474866331282, 0.07875025482159989, 0.08465798862251377, 
        0.09189094570840427, 0.10070853850346509, 0.11144193977503326, 
        0.124514471444123, 0.14046877432184984, 0.16000315768767867, 
        0.18402045973674444, 0.21369406779466352, 0.2505576217346058, 
        0.2966276062917161, 0.3545718989002555, 0.42794293056222643, 
        0.521502262206533, 0.6416753173042158, 0.7971926179830757, 1.0]
        >>> sa.prepare_filter_function(True, 0.04, False, 0.06, 1.5, 50, 0.05, False, None, 100) #doctest: +NORMALIZE_WHITESPACE
        [0.00083194331592229, 0.04086391086268285, 0.7994182049573009, 
        0.9554175714851911, 0.9257214133064412, 0.8864208155253327, 
        0.8406237416314447, 0.7895406073354527, 0.7344436719295907, 
        0.6766338461617276, 0.6173907887659099, 0.5579275104833195, 
        0.4993517885992762, 0.4426359119476192, 0.38859560599099563, 
        0.33787832130766693, 0.29096045886431027, 0.24815259496665634, 
        0.2096113871510978, 0.1753566038791111, 0.14529162554555938, 
        0.11922579924973033, 0.09689717267500671, 0.07799435496458187, 
        0.06217652402211632]
        >>> sa.prepare_filter_function(False, 0.04, False, 0.06, 1.0, 50, 0.05, False, None, 100) #doctest: +NORMALIZE_WHITESPACE
        [1.0, 0.9892079619944574, 0.9575256423365532, 0.9069606178873836, 
        0.8406237433345053, 0.7624126296654683, 0.676633846161729, 
        0.5876145767179657, 0.4993517885992762, 0.4152368286818413, 
        0.33787832130766693, 0.26902955708479037, 0.2096113871510978, 
        0.15981036888569505, 0.11922579924973033, 0.08703836765622351, 
        0.06217652402211632, 0.04346276456589661, 0.02972921638615875, 
        0.01989870361309264, 0.013032907448509368, 0.008352818518081014, 
        0.0052384160278331, 0.0032147124638421238, 0.0019304541362277093]
        """
        
        if low_pass_filter_option is True:
            freq_lowend = low_pass_filter_cutoff*pixelsize + filter_falloff/2
            low_fall_off = filter_falloff
        else:
            freq_lowend = 0.6
            low_fall_off = 1e-14
            
        if high_pass_filter_option is True:
            freq_highend = high_pass_filter_cutoff*pixelsize - filter_falloff/2
            high_fall_off = filter_falloff
        else:
            freq_highend = 0
            high_fall_off = 1e-14
            
        def compute_filter_value(frequency):
            filter_value = 0.5*(np.tanh(np.pi*(frequency + freq_highend)/(high_fall_off*(freq_highend - freq_lowend))) \
                 - np.tanh(np.pi*(frequency - freq_highend)/(high_fall_off*(freq_highend - freq_lowend)))\
                 - np.tanh(np.pi*(frequency + freq_lowend)/(low_fall_off*(freq_highend - freq_lowend))) \
                 + np.tanh(np.pi*(frequency - freq_lowend)/(low_fall_off*(freq_highend - freq_lowend))))
            return filter_value
        
        frequencies = [f for f in np.linspace(0, 0.5, image_dimension/2)]
        filter_function = [compute_filter_value(each_frequency) for each_frequency in frequencies]
        
        if custom_filter_option is True:
            custom_filter_function = self.read_custom_filter_file(custom_filter_file, image_dimension)
            filter_function = np.array(filter_function)*np.array(custom_filter_function)
        
        if bfactor != 0:
            fourier_coefficients = self.prepare_bfactor_coefficients(bfactor, pixelsize, image_dimension)
            filter_function = filter_function * fourier_coefficients / np.max(fourier_coefficients)
        
        if type(filter_function) != list:
            filter_function = filter_function.tolist()
            
        log_pairs = list(zip(frequencies, filter_function))
        msg = tabulate(log_pairs, ['normalized spatial frequencies', 'filter coefficients'])
        log_info = 'The following filter coefficients have been computed:\n{0}'.format(msg)
            
        self.log.ilog(log_info)
        
        return filter_function
    
    
    def read_custom_filter_file(self, custom_filter_file=None, image_dimension=None):
        """
        * Function to read Fourier coefficients from text file
        """
        if custom_filter_file is None: 
            custom_filter_file = self.custom_filter_file
        if image_dimension is None: 
            image_dimension = self.image_dimension
        
        ffile = open(custom_filter_file, 'r')
        filterrows = ffile.readlines()
        ffile.close()
        
        custom_filter_function = []
        for each_line in filterrows:
            if not each_line.startswith('#') or not each_line.startswith(';'):
                filter_pair = each_line.split()
                error = 'Last column entry filter coefficient could not be interpreted'
                try:
                    error += ' as a number' 
                    filter_coefficient = float(filter_pair[-1].strip())
                except:
                    raise ValueError(error)
                if filter_coefficient < 0 or filter_coefficient > 1:
                    error += ' between 0 and 1.' 
                custom_filter_function.append(filter_coefficient)
            
        if image_dimension/2 != len(custom_filter_function):
            error = """Filter file could not be interpreted. Number of detected rows does not match 
            image dimensions, i.e. Filter file should have 'image dimension/2' x 'frequency, 
            Fourier coefficient' value pairs"""
            raise ValueError(error)
        
        return custom_filter_function
        

    def log_mask_dimensions(self):
        
        return self.log.ilog('Maskfile                    : %s' % ('rectangle of {0} x {1} pixels'.\
        format(self.helixwidthpix, self.helixheightpix)))


    def prepare_mask(self, helixwidthpix, helixheightpix, image_dimension):
        mask = SegmentExam().make_smooth_rectangular_mask(helixwidthpix, helixheightpix, image_dimension)
        
        return mask
        
        
    def prepare_empty_rings(self, first_ring, last_ring, ring_step, full_circle_mode='F'):
        polar_interpolation_parameters = Numrinit(first_ring, last_ring, ring_step, full_circle_mode)
        
        ring_weights = ringwe(polar_interpolation_parameters, full_circle_mode)
        
        return polar_interpolation_parameters, ring_weights
        
        
    def center_and_rotate_image(self, ref_center, file_name, file_info, search_range_pix, delta_psi, y_align=False):
        img = EMData()
        img.read_image(file_name, file_info[0])
        segment_size = img.get_xsize()
        
        mask = SegmentExam().make_smooth_rectangular_mask(self.helixwidthpix, self.helixheightpix, segment_size)
        polar_interpolation_parameters, ring_weights = self.prepare_empty_rings(1, segment_size / 2 - 2, 1)
        
        cimage = self.generate_reference_rings_from_image(ref_center, polar_interpolation_parameters,
        ring_weights, segment_size)
        
        reference_rings = [cimage]
        
        centered_file = os.path.join(self.tempdir, os.path.basename(file_name))
        x_range = search_range_pix
        if not y_align:
            y_range = 0
        else: 
            y_range = x_range
            
        step_x = 0.3
        full_circle_mode = 'F'
        center_x = center_y = segment_size / 2 + 1
        previous_shift_x = previous_shift_y = previous_inplane_angle = 0.0
        
        for each_index, each_image in enumerate(file_info):
            img.read_image(file_name, each_image)
            img *= mask
            
            [inplane_angle, sxst, syst, mirror, matched_reference_id, peak] = Util.multiref_polar_ali_2d_delta(img,
            reference_rings, x_range, y_range, step_x, full_circle_mode, polar_interpolation_parameters, center_x -
            previous_shift_x, center_y - previous_shift_y, previous_inplane_angle, delta_psi)
            
            angb, sxb, syb, ct = compose_transform2(0.0, sxst, syst, 1, -inplane_angle, 0.0, 0.0, 1)
            shift_x = previous_shift_x - sxb
            shift_y = previous_shift_y - syb
            
            centered_img = rot_shift2D(img, inplane_angle, shift_x, shift_y)
            centered_img.write_image(centered_file, each_index)
            
        return centered_file


    def prepare_reference_stack(self, reference_stack_name):
        blank_image = EMData()
        blank_image.read_image(reference_stack_name, 0)
        
        blank_image.to_zero()
        
        reference_stack = []
        reference_features = namedtuple('reference_features', 
        'odd_average even_average total_average variance number_of_images')
        
        references_image_count = EMUtil.get_image_count(reference_stack_name)
        ref_image = EMData()
        for each_ref_id in list(range(references_image_count)):
            ref_image.read_image(reference_stack_name, each_ref_id)
            number_of_images = defaultdict(int)
            
            reference_imgs = reference_features(blank_image.copy(), blank_image.copy(), ref_image.copy(),
            blank_image.copy(), number_of_images)
            
            reference_stack.append(reference_imgs)
        
        return reference_stack


    def average_stack(self, stack, img_info, align=False):
        img = EMData()
        img.read_image(stack)
        
        avg = model_blank(img.get_xsize(), img.get_ysize(), 1, 0)
        for each_image in img_info:
            img.read_image(stack, each_image)
            if align and img.has_attr('xform.align2d'):
                t = img.get_attr('xform.align2d')
                d = t.get_params('2D')
                img = rot_shift2D(img, d['alpha'], d['tx'], d['ty'])
                
            avg += img
            
        avg /= len(img_info)
        
        return avg
        

    def center_reference_images_by_alignment_to_avg(self, file_info, ref_center):
        search_range_pix = int(50 / self.pixelsize)
        
        self.reference_stack_name = \
        self.center_and_rotate_image(ref_center, self.reference_stack_name, file_info, search_range_pix, self.delta_psi)
        
        return self.reference_stack_name

    
    def get_image_list_named_tuple(self):
        return namedtuple('alignments', 'stack_id local_id ref_id shift_x shift_y inplane_angle peak mirror')
    

    def get_image_alignments_from_stack(self, alignment_stack_name):
        stack_image_count = EMUtil.get_image_count(alignment_stack_name)
        image_nt = self.get_image_list_named_tuple()
        img = EMData()
        images_info = []
        for each_local_id, each_image_id in enumerate(list(range(stack_image_count))):
            img.read_image(alignment_stack_name, each_image_id)
            if img.has_attr('xform.align2d') and self.refine_locally:
                inplane_angle, shift_x, shift_y, mirror, scale = get_params2D(img)
                
                inplane_angle, shift_x, shift_y = Segment().convert_rotate_shift_to_shift_rotate_order(inplane_angle,
                shift_x, shift_y)
                
                image_info = image_nt(each_image_id, each_local_id, 0, -shift_x * self.pixelsize, 
                    -shift_y * self.pixelsize, inplane_angle, 0, mirror)
            else:
                image_info = image_nt(each_image_id, each_local_id, 0, 0, 0, 0, 0, 0)
            images_info.append(image_info)
        
        return images_info, img.get_xsize()


    def compute_binfactor_for_desired_resolution(self, desired_resolution, pixelsize):
        """
        >>> from spring.segment2d.segmentalign2d import SegmentAlign2d
        >>> s = SegmentAlign2d()
        >>> resolutions = np.array([24.0, 12.0, 7.0, 3.0])
        >>> s.compute_binfactor_for_desired_resolution(resolutions, 2.4)
        array([3, 2, 1])
        >>> s.compute_binfactor_for_desired_resolution(resolutions, 1.2)
        array([7, 3, 2, 1])
        >>> s.compute_binfactor_for_desired_resolution(resolutions, 1.372)
        array([6, 3, 2, 1])
        """
        binfactor = np.array(np.round(desired_resolution / (3 * pixelsize)), dtype=int)
        binfactor = binfactor[binfactor >= 1]
        
        return binfactor
    

    def get_align_info_nt(self):
        align_info = namedtuple('align_info', 'iteration_id pixelsize binfactor x_range y_range')

        return align_info


    def define_flow_of_alignment(self, pixelsize, binfactor, iteration_count, x_range_A, y_range_A):
        """
        >>> from spring.segment2d.segmentalign2d import SegmentAlign2d
        >>> SegmentAlign2d().define_flow_of_alignment(1.2, 2, 2, 50, 50) #doctest: +NORMALIZE_WHITESPACE
        [align_info(iteration_id=0, pixelsize=8.4000000000000004, binfactor=7, x_range=5.9523809523809526, y_range=5.9523809523809526), 
        align_info(iteration_id=1, pixelsize=8.4000000000000004, binfactor=7, x_range=5.9523809523809526, y_range=5.9523809523809526), 
        align_info(iteration_id=0, pixelsize=3.5999999999999996, binfactor=3, x_range=5.9523809523809526, y_range=5.9523809523809526), 
        align_info(iteration_id=1, pixelsize=3.5999999999999996, binfactor=3, x_range=5.9523809523809526, y_range=5.9523809523809526), 
        align_info(iteration_id=0, pixelsize=2.3999999999999999, binfactor=2, x_range=5.9523809523809526, y_range=5.9523809523809526), 
        align_info(iteration_id=1, pixelsize=2.3999999999999999, binfactor=2, x_range=5.9523809523809526, y_range=5.9523809523809526)]
        
        >>> SegmentAlign2d().define_flow_of_alignment(5.0, 2, 5, 50, 50) #doctest: +NORMALIZE_WHITESPACE
        [align_info(iteration_id=0, pixelsize=10.0, binfactor=2, x_range=5.0, y_range=5.0), 
        align_info(iteration_id=1, pixelsize=10.0, binfactor=2, x_range=5.0, y_range=5.0), 
        align_info(iteration_id=2, pixelsize=10.0, binfactor=2, x_range=5.0, y_range=5.0), 
        align_info(iteration_id=3, pixelsize=10.0, binfactor=2, x_range=5.0, y_range=5.0), 
        align_info(iteration_id=4, pixelsize=10.0, binfactor=2, x_range=5.0, y_range=5.0)]
        
        >>> SegmentAlign2d().define_flow_of_alignment(5.0, 1, 5, 50, 50) #doctest: +NORMALIZE_WHITESPACE
        [align_info(iteration_id=0, pixelsize=10.0, binfactor=2, x_range=5.0, y_range=5.0), 
        align_info(iteration_id=1, pixelsize=10.0, binfactor=2, x_range=5.0, y_range=5.0), 
        align_info(iteration_id=2, pixelsize=10.0, binfactor=2, x_range=5.0, y_range=5.0), 
        align_info(iteration_id=3, pixelsize=10.0, binfactor=2, x_range=5.0, y_range=5.0), 
        align_info(iteration_id=4, pixelsize=10.0, binfactor=2, x_range=5.0, y_range=5.0), 
        align_info(iteration_id=0, pixelsize=5.0, binfactor=1, x_range=5.0, y_range=5.0), 
        align_info(iteration_id=1, pixelsize=5.0, binfactor=1, x_range=5.0, y_range=5.0), 
        align_info(iteration_id=2, pixelsize=5.0, binfactor=1, x_range=5.0, y_range=5.0), 
        align_info(iteration_id=3, pixelsize=5.0, binfactor=1, x_range=5.0, y_range=5.0), 
        align_info(iteration_id=4, pixelsize=5.0, binfactor=1, x_range=5.0, y_range=5.0)]
        """
        align_info = self.get_align_info_nt()
        
        alignment_info = []
        bin_factor_series = self.compute_binfactor_for_desired_resolution(np.array([24.0, 12.0, 7.0, 3.0]), pixelsize)
        bin_factor_series = bin_factor_series[bin_factor_series >= binfactor]
        
        x_range = x_range_A / (max(bin_factor_series) * pixelsize)
        y_range = y_range_A / (max(bin_factor_series) * pixelsize)

        for each_bin_factor in bin_factor_series:
            for each_iteration in list(range(iteration_count)):
                alignment_info.append(align_info(each_iteration, each_bin_factor * pixelsize, each_bin_factor, x_range,
                y_range))
        
        self.log.ilog('The following flow of alignment has been determined: {0}'.format(alignment_info.__str__()))
        
        return alignment_info
    
    
    def prepare_alignment(self):
        OpenMpi().setup_and_start_mpi_version_if_demanded(self.mpi_option, self.feature_set, self.cpu_count)
        self.tempdir = Temporary().mktmpdir(self.temppath)
            
        self.log.fcttolog()
        
        images_info, self.image_dimension = self.get_image_alignments_from_stack(self.alignment_stack_name)
        
        if not self.reference_option:
            image_ids = [each_image.local_id for each_image in images_info]
            ref_center = self.average_stack(self.alignment_stack_name, image_ids, align=True)
            ref_center.write_image('average.hdf')
        
            self.reference_stack_name = 'average.hdf'
            
        alignment_info = self.define_flow_of_alignment(self.pixelsize, self.binfactor, self.number_of_iterations,
        self.x_range_A, self.y_range_A)
        
        self.log.plog(10)
        
        return alignment_info, images_info
        